from enum import Enum

from pydantic import Field
from nodetool.metadata.types import ImageTensor
from nodetool.common.comfy_node import DensePoseModel

from nodetool.common.comfy_node import ComfyNode, EnableDisable
from nodetool.nodes.comfy.controlnet import PreprocessImage


class Toggle(str, Enum):
    ENABLE = "enable"
    DISABLE = "disable"


class DensePoseCMap(str, Enum):
    VIRIDIS = "Viridis (MagicAnimate)"
    PARULA = "Parula (CivitAI)"


class BBoxDetectorModel(str, Enum):
    YOLOX_L_TORCHSCRIPT_PT = "yolox_l.torchscript.pt"
    YOLOX_L_ONNX = "yolox_l.onnx"
    YOLO_NAS_L_FP16_ONNX = "yolo_nas_l_fp16.onnx"
    YOLO_NAS_M_FP16_ONNX = "yolo_nas_m_fp16.onnx"
    YOLO_NAS_S_FP16_ONNX = "yolo_nas_s_fp16.onnx"


class PoseEstimatorModel(str, Enum):
    DW_LL_UCOCO_384_BS5_TORCHSCRIPT_PT = "dw-ll_ucoco_384_bs5.torchscript.pt"
    DW_LL_UCOCO_384_ONNX = "dw-ll_ucoco_384.onnx"
    DW_LL_UCOCO_ONNX = "dw-ll_ucoco.onnx"


class DensePosePreprocessor(PreprocessImage):
    model: DensePoseModel = Field(
        default=DensePoseModel.DENSEPOSE_R50_FPN_DL, description="The model to use."
    )
    cmap: DensePoseCMap = Field(
        default=DensePoseCMap.VIRIDIS, description="The color map to use."
    )


class DWPose_Preprocessor(ComfyNode):
    detect_hand: Toggle = Field(
        default=Toggle.ENABLE, description="Toggle to enable or disable hand detection."
    )
    detect_body: Toggle = Field(
        default=Toggle.ENABLE, description="Toggle to enable or disable body detection."
    )
    detect_face: Toggle = Field(
        default=Toggle.ENABLE, description="Toggle to enable or disable face detection."
    )
    bbox_detector: BBoxDetectorModel = Field(
        default=BBoxDetectorModel.YOLOX_L_TORCHSCRIPT_PT,
        description="The bounding box detector model to use.",
    )
    pose_estimator: PoseEstimatorModel = Field(
        default=PoseEstimatorModel.DW_LL_UCOCO_384_BS5_TORCHSCRIPT_PT,
        description="The pose estimator model to use.",
    )

    @classmethod
    def return_type(cls):
        return {"type": ImageTensor}


class OpenposePreprocessor(PreprocessImage):
    detect_hand: EnableDisable = Field(
        default=EnableDisable.ENABLE,
        description="Whether to detect hands.",
    )
    detect_body: EnableDisable = Field(
        default=EnableDisable.ENABLE,
        description="Whether to detect bodies.",
    )
    detect_face: EnableDisable = Field(
        default=EnableDisable.ENABLE,
        description="Whether to detect faces.",
    )


class MediaPipeFaceMeshPreprocessor(PreprocessImage):
    max_faces: int = Field(
        default=10, description="The maximum number of faces to detect."
    )
    min_confidence: float = Field(
        default=0.5, description="The minimum confidence for face detection."
    )



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class CropMask(GraphNode):
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to crop.')
    x: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The x position for cropping.')
    y: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The y position for cropping.')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Width of the crop.')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Height of the crop.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.CropMask"



class FeatherMask(GraphNode):
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to feather.')
    left: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Feather amount on the left.')
    top: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Feather amount on the top.')
    right: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Feather amount on the right.')
    bottom: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='Feather amount on the bottom.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.FeatherMask"



class GrowMask(GraphNode):
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to grow.')
    expand: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The amount to expand the mask.')
    tapered_corners: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Whether to taper the corners.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.GrowMask"



class ImageColorToMask(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to extract the color mask.')
    color: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The color to use for the mask.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.ImageColorToMask"


from nodetool.nodes.comfy.mask import ChannelEnum

class ImageToMask(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to extract the mask.')
    channel: ChannelEnum | GraphNode | tuple[GraphNode, str] = Field(default=ChannelEnum('red'), description='The channel to use for the mask.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.ImageToMask"



class InvertMask(GraphNode):
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to invert.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.InvertMask"


from nodetool.nodes.comfy.mask import OperationEnum

class MaskComposite(GraphNode):
    destination: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The destination mask.')
    source: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The source mask.')
    x: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The x position.')
    y: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The y position.')
    operation: OperationEnum | GraphNode | tuple[GraphNode, str] = Field(default=OperationEnum('multiply'), description='The operation to use.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.MaskComposite"



class MaskToImage(GraphNode):
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to convert.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.MaskToImage"



class SolidMask(GraphNode):
    value: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The value for the solid mask.')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Width of the solid mask.')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Height of the solid mask.')
    @classmethod
    def get_node_type(cls): return "comfy.mask.SolidMask"





# -*- coding: utf-8 -*-
# Copyright (c) Facebook, Inc. and its affiliates.
from custom_detectron2.config import CfgNode as CN

__all__ = ["add_common_config", "add_oneformer_config", "add_swin_config", 
            "add_dinat_config", "add_beit_adapter_config", "add_convnext_config"]

def add_common_config(cfg):
    """
    Add config for common configuration
    """
    # data config
    # select the dataset mapper
    cfg.INPUT.DATASET_MAPPER_NAME = "oneformer_unified"
    # Color augmentation
    cfg.INPUT.COLOR_AUG_SSD = False
    # We retry random cropping until no single category in semantic segmentation GT occupies more
    # than `SINGLE_CATEGORY_MAX_AREA` part of the crop.
    cfg.INPUT.CROP.SINGLE_CATEGORY_MAX_AREA = 1.0
    # Pad image and segmentation GT in dataset mapper.
    cfg.INPUT.SIZE_DIVISIBILITY = -1

    cfg.INPUT.TASK_SEQ_LEN = 77
    cfg.INPUT.MAX_SEQ_LEN = 77

    cfg.INPUT.TASK_PROB = CN()
    cfg.INPUT.TASK_PROB.SEMANTIC = 0.33
    cfg.INPUT.TASK_PROB.INSTANCE = 0.66

    # test dataset
    cfg.DATASETS.TEST_PANOPTIC = ("",)
    cfg.DATASETS.TEST_INSTANCE = ("",)
    cfg.DATASETS.TEST_SEMANTIC = ("",)

    # solver config
    # weight decay on embedding
    cfg.SOLVER.WEIGHT_DECAY_EMBED = 0.0
    # optimizer
    cfg.SOLVER.OPTIMIZER = "ADAMW"
    cfg.SOLVER.BACKBONE_MULTIPLIER = 0.1

    # wandb
    cfg.WANDB = CN()
    cfg.WANDB.PROJECT = "unified_dense_recognition"
    cfg.WANDB.NAME = None

    cfg.MODEL.IS_TRAIN = False
    cfg.MODEL.IS_DEMO = True

    # text encoder config
    cfg.MODEL.TEXT_ENCODER = CN()

    cfg.MODEL.TEXT_ENCODER.WIDTH = 256
    cfg.MODEL.TEXT_ENCODER.CONTEXT_LENGTH = 77
    cfg.MODEL.TEXT_ENCODER.NUM_LAYERS = 12
    cfg.MODEL.TEXT_ENCODER.VOCAB_SIZE = 49408
    cfg.MODEL.TEXT_ENCODER.PROJ_NUM_LAYERS = 2
    cfg.MODEL.TEXT_ENCODER.N_CTX = 16

    # mask_former inference config
    cfg.MODEL.TEST = CN()
    cfg.MODEL.TEST.SEMANTIC_ON = True
    cfg.MODEL.TEST.INSTANCE_ON = False
    cfg.MODEL.TEST.PANOPTIC_ON = False
    cfg.MODEL.TEST.DETECTION_ON = False
    cfg.MODEL.TEST.OBJECT_MASK_THRESHOLD = 0.0
    cfg.MODEL.TEST.OVERLAP_THRESHOLD = 0.0
    cfg.MODEL.TEST.SEM_SEG_POSTPROCESSING_BEFORE_INFERENCE = False
    cfg.MODEL.TEST.TASK = "panoptic"

    # TEST AUG Slide
    cfg.TEST.AUG.IS_SLIDE = False
    cfg.TEST.AUG.CROP_SIZE = (640, 640)
    cfg.TEST.AUG.STRIDE = (426, 426)
    cfg.TEST.AUG.SCALE = (2048, 640)
    cfg.TEST.AUG.SETR_MULTI_SCALE = True
    cfg.TEST.AUG.KEEP_RATIO = True
    cfg.TEST.AUG.SIZE_DIVISOR = 32

    # pixel decoder config
    cfg.MODEL.SEM_SEG_HEAD.MASK_DIM = 256
    # adding transformer in pixel decoder
    cfg.MODEL.SEM_SEG_HEAD.TRANSFORMER_ENC_LAYERS = 0
    # pixel decoder
    cfg.MODEL.SEM_SEG_HEAD.PIXEL_DECODER_NAME = "BasePixelDecoder"
    cfg.MODEL.SEM_SEG_HEAD.SEM_EMBED_DIM = 256
    cfg.MODEL.SEM_SEG_HEAD.INST_EMBED_DIM = 256

    # LSJ aug
    cfg.INPUT.IMAGE_SIZE = 1024
    cfg.INPUT.MIN_SCALE = 0.1
    cfg.INPUT.MAX_SCALE = 2.0

    # MSDeformAttn encoder configs
    cfg.MODEL.SEM_SEG_HEAD.DEFORMABLE_TRANSFORMER_ENCODER_IN_FEATURES = ["res3", "res4", "res5"]
    cfg.MODEL.SEM_SEG_HEAD.DEFORMABLE_TRANSFORMER_ENCODER_N_POINTS = 4
    cfg.MODEL.SEM_SEG_HEAD.DEFORMABLE_TRANSFORMER_ENCODER_N_HEADS = 8

def add_oneformer_config(cfg):
    """
    Add config for ONE_FORMER.
    """

    # mask_former model config
    cfg.MODEL.ONE_FORMER = CN()

    # loss
    cfg.MODEL.ONE_FORMER.DEEP_SUPERVISION = True
    cfg.MODEL.ONE_FORMER.NO_OBJECT_WEIGHT = 0.1
    cfg.MODEL.ONE_FORMER.CLASS_WEIGHT = 1.0
    cfg.MODEL.ONE_FORMER.DICE_WEIGHT = 1.0
    cfg.MODEL.ONE_FORMER.MASK_WEIGHT = 20.0
    cfg.MODEL.ONE_FORMER.CONTRASTIVE_WEIGHT = 0.5
    cfg.MODEL.ONE_FORMER.CONTRASTIVE_TEMPERATURE = 0.07

    # transformer config
    cfg.MODEL.ONE_FORMER.NHEADS = 8
    cfg.MODEL.ONE_FORMER.DROPOUT = 0.1
    cfg.MODEL.ONE_FORMER.DIM_FEEDFORWARD = 2048
    cfg.MODEL.ONE_FORMER.ENC_LAYERS = 0
    cfg.MODEL.ONE_FORMER.CLASS_DEC_LAYERS = 2
    cfg.MODEL.ONE_FORMER.DEC_LAYERS = 6
    cfg.MODEL.ONE_FORMER.PRE_NORM = False

    cfg.MODEL.ONE_FORMER.HIDDEN_DIM = 256
    cfg.MODEL.ONE_FORMER.NUM_OBJECT_QUERIES = 120
    cfg.MODEL.ONE_FORMER.NUM_OBJECT_CTX = 16
    cfg.MODEL.ONE_FORMER.USE_TASK_NORM = True

    cfg.MODEL.ONE_FORMER.TRANSFORMER_IN_FEATURE = "res5"
    cfg.MODEL.ONE_FORMER.ENFORCE_INPUT_PROJ = False

    # Sometimes `backbone.size_divisibility` is set to 0 for some backbone (e.g. ResNet)
    # you can use this config to override
    cfg.MODEL.ONE_FORMER.SIZE_DIVISIBILITY = 32

    # transformer module
    cfg.MODEL.ONE_FORMER.TRANSFORMER_DECODER_NAME = "ContrastiveMultiScaleMaskedTransformerDecoder"

    # point loss configs
    # Number of points sampled during training for a mask point head.
    cfg.MODEL.ONE_FORMER.TRAIN_NUM_POINTS = 112 * 112
    # Oversampling parameter for PointRend point sampling during training. Parameter `k` in the
    # original paper.
    cfg.MODEL.ONE_FORMER.OVERSAMPLE_RATIO = 3.0
    # Importance sampling parameter for PointRend point sampling during training. Parametr `beta` in
    # the original paper.
    cfg.MODEL.ONE_FORMER.IMPORTANCE_SAMPLE_RATIO = 0.75

def add_swin_config(cfg):
    """
    Add config forSWIN Backbone.
    """
    
    # swin transformer backbone
    cfg.MODEL.SWIN = CN()
    cfg.MODEL.SWIN.PRETRAIN_IMG_SIZE = 224
    cfg.MODEL.SWIN.PATCH_SIZE = 4
    cfg.MODEL.SWIN.EMBED_DIM = 96
    cfg.MODEL.SWIN.DEPTHS = [2, 2, 6, 2]
    cfg.MODEL.SWIN.NUM_HEADS = [3, 6, 12, 24]
    cfg.MODEL.SWIN.WINDOW_SIZE = 7
    cfg.MODEL.SWIN.MLP_RATIO = 4.0
    cfg.MODEL.SWIN.QKV_BIAS = True
    cfg.MODEL.SWIN.QK_SCALE = None
    cfg.MODEL.SWIN.DROP_RATE = 0.0
    cfg.MODEL.SWIN.ATTN_DROP_RATE = 0.0
    cfg.MODEL.SWIN.DROP_PATH_RATE = 0.3
    cfg.MODEL.SWIN.APE = False
    cfg.MODEL.SWIN.PATCH_NORM = True
    cfg.MODEL.SWIN.OUT_FEATURES = ["res2", "res3", "res4", "res5"]
    cfg.MODEL.SWIN.USE_CHECKPOINT = False
    ## Semask additions
    cfg.MODEL.SWIN.SEM_WINDOW_SIZE = 7
    cfg.MODEL.SWIN.NUM_SEM_BLOCKS = 1

def add_dinat_config(cfg):
    """
    Add config for NAT Backbone.
    """

    # DINAT transformer backbone
    cfg.MODEL.DiNAT = CN()
    cfg.MODEL.DiNAT.DEPTHS = [3, 4, 18, 5]
    cfg.MODEL.DiNAT.OUT_FEATURES = ["res2", "res3", "res4", "res5"]
    cfg.MODEL.DiNAT.EMBED_DIM = 64
    cfg.MODEL.DiNAT.MLP_RATIO = 3.0
    cfg.MODEL.DiNAT.NUM_HEADS = [2, 4, 8, 16]
    cfg.MODEL.DiNAT.DROP_PATH_RATE = 0.2
    cfg.MODEL.DiNAT.KERNEL_SIZE = 7
    cfg.MODEL.DiNAT.DILATIONS = [[1, 16, 1], [1, 4, 1, 8], [1, 2, 1, 3, 1, 4], [1, 2, 1, 2, 1]]
    cfg.MODEL.DiNAT.OUT_INDICES = (0, 1, 2, 3)
    cfg.MODEL.DiNAT.QKV_BIAS = True
    cfg.MODEL.DiNAT.QK_SCALE = None
    cfg.MODEL.DiNAT.DROP_RATE = 0
    cfg.MODEL.DiNAT.ATTN_DROP_RATE = 0.
    cfg.MODEL.DiNAT.IN_PATCH_SIZE = 4

def add_convnext_config(cfg):
    """
    Add config for ConvNeXt Backbone.
    """
    
    # swin transformer backbone
    cfg.MODEL.CONVNEXT = CN()
    cfg.MODEL.CONVNEXT.IN_CHANNELS = 3
    cfg.MODEL.CONVNEXT.DEPTHS = [3, 3, 27, 3]
    cfg.MODEL.CONVNEXT.DIMS = [192, 384, 768, 1536]
    cfg.MODEL.CONVNEXT.DROP_PATH_RATE = 0.4
    cfg.MODEL.CONVNEXT.LSIT = 1.0
    cfg.MODEL.CONVNEXT.OUT_INDICES = [0, 1, 2, 3]
    cfg.MODEL.CONVNEXT.OUT_FEATURES = ["res2", "res3", "res4", "res5"]

def add_beit_adapter_config(cfg):
    """
    Add config for BEiT Adapter Backbone.
    """

    # beit adapter backbone
    cfg.MODEL.BEiTAdapter = CN()
    cfg.MODEL.BEiTAdapter.IMG_SIZE = 640
    cfg.MODEL.BEiTAdapter.PATCH_SIZE = 16
    cfg.MODEL.BEiTAdapter.EMBED_DIM = 1024
    cfg.MODEL.BEiTAdapter.DEPTH = 24
    cfg.MODEL.BEiTAdapter.NUM_HEADS = 16
    cfg.MODEL.BEiTAdapter.MLP_RATIO = 4
    cfg.MODEL.BEiTAdapter.QKV_BIAS = True
    cfg.MODEL.BEiTAdapter.USE_ABS_POS_EMB = False
    cfg.MODEL.BEiTAdapter.USE_REL_POS_BIAS = True
    cfg.MODEL.BEiTAdapter.INIT_VALUES = 1e-6
    cfg.MODEL.BEiTAdapter.DROP_PATH_RATE = 0.3
    cfg.MODEL.BEiTAdapter.CONV_INPLANE = 64
    cfg.MODEL.BEiTAdapter.N_POINTS = 4
    cfg.MODEL.BEiTAdapter.DEFORM_NUM_HEADS = 16
    cfg.MODEL.BEiTAdapter.CFFN_RATIO = 0.25
    cfg.MODEL.BEiTAdapter.DEFORM_RATIO = 0.5
    cfg.MODEL.BEiTAdapter.WITH_CP = True
    cfg.MODEL.BEiTAdapter.INTERACTION_INDEXES=[[0, 5], [6, 11], [12, 17], [18, 23]]
    cfg.MODEL.BEiTAdapter.OUT_FEATURES = ["res2", "res3", "res4", "res5"]


from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management
import numpy as np

class MLSD_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            score_threshold = ("FLOAT", {"default": 0.1, "min": 0.01, "max": 2.0, "step": 0.01}),
            dist_threshold = ("FLOAT", {"default": 0.1, "min": 0.01, "max": 20.0, "step": 0.01})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, score_threshold, dist_threshold, resolution=512, **kwargs):
        from controlnet_aux.mlsd import MLSDdetector

        model = MLSDdetector.from_pretrained().to(model_management.get_torch_device())
        out = common_annotator_call(model, image, resolution=resolution, thr_v=score_threshold, thr_d=dist_threshold)
        return (out, )

NODE_CLASS_MAPPINGS = {
    "M-LSDPreprocessor": MLSD_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "M-LSDPreprocessor": "M-LSD Lines"
}


from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class LineArt_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            coarse=(["disable", "enable"], {"default": "disable"})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, resolution=512, **kwargs):
        from controlnet_aux.lineart import LineartDetector

        model = LineartDetector.from_pretrained().to(model_management.get_torch_device())
        out = common_annotator_call(model, image, resolution=resolution, coarse = kwargs["coarse"] == "enable")
        del model
        return (out, )

NODE_CLASS_MAPPINGS = {
    "LineArtPreprocessor": LineArt_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "LineArtPreprocessor": "Realistic Lineart"
}


# Copyright (c) Facebook, Inc. and its affiliates.
from . import data  # register all new datasets
from . import modeling

# config
from .config import *

# models
from .oneformer_model import OneFormer


from pydantic import Field
from nodetool.workflows.base_node import BaseNode


class HTTPGet(BaseNode):
    """
    Make a HTTP GET request to a URL.
    http, get, request, url
    """

    url: str = Field(
        default="",
        description="The URL to make the GET request to.",
    )

    async def process(self, context) -> str:
        async with context.http_session.get(self.url) as response:
            return await response.text()


class HTTPPost(BaseNode):
    """
    Make a HTTP POST request to a URL.
    http, post, request, url
    """

    url: str = Field(
        default="",
        description="The URL to make the POST request to.",
    )
    data: str = Field(
        default="",
        description="The data to send in the POST request.",
    )

    async def process(self, context) -> str:
        async with context.http_session.post(self.url, data=self.data) as response:
            return await response.text()


class HTTPPut(BaseNode):
    """
    Make a HTTP PUT request to a URL.
    http, put, request, url
    """

    url: str = Field(
        default="",
        description="The URL to make the PUT request to.",
    )
    data: str = Field(
        default="",
        description="The data to send in the PUT request.",
    )

    async def process(self, context) -> str:
        async with context.http_session.put(self.url, data=self.data) as response:
            return await response.text()


class HTTPDelete(BaseNode):
    """
    Make a HTTP DELETE request to a URL.
    http, delete, request, url
    """

    url: str = Field(
        default="",
        description="The URL to make the DELETE request to.",
    )

    async def process(self, context) -> str:
        async with context.http_session.delete(self.url) as response:
            return await response.text()



from datetime import datetime
import uuid
from typing import Any, Optional
from nodetool.api.types.graph import Graph as APIGraph
from nodetool.workflows.graph import Graph
from nodetool.workflows.base_node import BaseNode

from nodetool.models.base_model import DBModel, DBField, create_time_ordered_uuid


class Workflow(DBModel):

    @classmethod
    def get_table_schema(cls):
        return {
            "table_name": "nodetool_workflows",
            "key_schema": {"id": "HASH"},
            "attribute_definitions": {
                "id": "S",
                "user_id": "S",
                "access": "S",
                "updated_at": "S",
            },
            "global_secondary_indexes": {
                "nodetool_workflow_user_index": {"user_id": "HASH"},
                "nodetool_workflow_access_index": {
                    "access": "HASH",
                    "updated_at": "RANGE",
                },
            },
        }

    id: str = DBField(hash_key=True)
    user_id: str = DBField(default="")
    access: str = DBField(default="private")
    created_at: datetime = DBField(default_factory=datetime.now)
    updated_at: datetime = DBField(default_factory=datetime.now)
    name: str = DBField(default="")
    description: str | None = DBField(default="")
    thumbnail: str | None = DBField(default=None)
    graph: dict = DBField(default_factory=dict)

    def before_save(self):
        self.updated_at = datetime.now()

    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        """
        Create a new Workflow object from a dictionary.
        """
        return cls(
            id=data["id"],
            user_id=data["user_id"],
            access=data["access"],
            created_at=data["created_at"],
            updated_at=data["updated_at"],
            name=data["name"],
            description=data["description"],
            thumbnail=data["thumbnail"],
            graph=data["graph"],
        )

    @classmethod
    def find(cls, user_id: str, workflow_id: str):
        workflow = cls.get(workflow_id)
        return (
            workflow
            if workflow and (workflow.user_id == user_id or workflow.access == "public")
            else None
        )

    @classmethod
    def create(cls, user_id: str, name: str, graph: dict[str, Any], **kwargs):
        """
        Create a new image in the database.
        """

        return super().create(
            id=create_time_ordered_uuid(),
            user_id=user_id,
            name=name,
            graph=graph,
            **kwargs,
        )

    @classmethod
    def paginate(
        cls,
        user_id: str | None = None,
        limit: int = 100,
        start_key: Optional[str] = None,
    ):
        if user_id is None:
            return cls.query(
                condition="access = :public",
                values={":public": "public"},
                index="nodetool_workflow_access_index",
                limit=limit,
                start_key=start_key,
            )
        else:
            return cls.query(
                condition="user_id = :user_id",
                values={":user_id": user_id},
                index="nodetool_workflow_user_index",
                limit=limit,
                start_key=start_key,
            )

    def get_api_graph(self) -> APIGraph:
        """
        Returns the graph object for the workflow.
        """
        return APIGraph(
            nodes=self.graph["nodes"],
            edges=self.graph["edges"],
        )

    def get_graph(self) -> Graph:
        """
        Returns the graph object for the workflow.
        """
        return Graph(
            nodes=[
                BaseNode.from_dict(node, skip_errors=True)
                for node in self.graph["nodes"]
            ],
            edges=self.graph["edges"],
        )



import nodetool.nodes.huggingface.audio.generate
import nodetool.nodes.huggingface.image.generate
import nodetool.nodes.huggingface.image.segment
import nodetool.nodes.huggingface.image.classify
import nodetool.nodes.huggingface.text.classify
import nodetool.nodes.huggingface.text.summarize
import nodetool.nodes.huggingface.text.generate



#!/usr/bin/env python

import asyncio
from datetime import datetime
import json
import threading
from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import StreamingResponse
from nodetool.api.utils import current_user, User

from nodetool.api.types.job import (
    JobList,
    JobUpdate,
)
from nodetool.workflows.run_job_request import RunJobRequest
from nodetool.common.environment import Environment

from nodetool.models.job import Job
from nodetool.models.workflow import Workflow
from nodetool.models.prediction import Prediction
from nodetool.api.types.prediction import Prediction as APIPrediction
from nodetool.workflows.types import Error, WorkflowUpdate


log = Environment.get_logger()
router = APIRouter(prefix="/api/jobs", tags=["jobs"])


@router.get("/{id}")
async def get(id: str, user: User = Depends(current_user)) -> Job:
    """
    Returns the status of a job.
    """
    job = Job.find(user.id, id)
    if job is None:
        raise HTTPException(status_code=404, detail="Job not found")
    else:
        if job.user_id != user.id:
            raise HTTPException(status_code=403, detail="Forbidden")
        else:
            return job


@router.get("/")
async def index(
    workflow_id: str | None = None,
    cursor: str | None = None,
    page_size: int | None = None,
    user: User = Depends(current_user),
) -> JobList:
    """
    Returns all assets for a given user or workflow.
    """
    if page_size is None:
        page_size = 10

    jobs, next_cursor = Job.paginate(
        user_id=user.id, workflow_id=workflow_id, limit=page_size, start_key=cursor
    )

    return JobList(next=next_cursor, jobs=jobs)


@router.put("/{id}")
async def update(id: str, req: JobUpdate, user: User = Depends(current_user)) -> Job:
    """
    Update a job.
    """
    job = Job.find(user.id, id)
    if job is None:
        raise HTTPException(status_code=404, detail="Job not found")
    else:
        if job.user_id != user.id:
            raise HTTPException(status_code=403, detail="Forbidden")
        else:
            job.status = req.status
            job.error = req.error
            job.save()
            return job


@router.post("/")
async def run(
    req: RunJobRequest, execute: bool = True, user: User = Depends(current_user)
):
    from nodetool.workflows.workflow_runner import WorkflowRunner
    from nodetool.workflows.processing_context import (
        ProcessingContext,
    )

    if req.graph is None:
        workflow = Workflow.find(user.id, req.workflow_id)
        if workflow is None:
            raise HTTPException(status_code=404, detail="Workflow not found")
        else:
            req.graph = workflow.get_api_graph()

    assert req.graph is not None, "Graph is required"

    job = Job.create(
        job_type=req.job_type,
        workflow_id=req.workflow_id,
        user_id=user.id,
        graph=req.graph.model_dump(),
        status="running",
    )

    if execute is False:
        return job

    assert user.auth_token

    context = ProcessingContext(
        user_id=user.id,
        auth_token=user.auth_token,
        workflow_id=req.workflow_id,
    )

    runner = WorkflowRunner()

    async def run():
        try:
            await runner.run(req, context)
        except Exception as e:
            log.exception(e)
            context.post_message(Error(error=str(e)))

    async def generate():
        yield json.dumps(JobUpdate(status="running").model_dump()) + "\n"

        thread = threading.Thread(target=lambda: asyncio.run(run()))
        thread.start()
        try:
            while runner.is_running():
                if context.has_messages():
                    msg = await context.pop_message_async()
                    if isinstance(msg, Prediction):
                        msg = APIPrediction.from_model(msg)
                    if isinstance(msg, WorkflowUpdate):
                        job.finished_at = datetime.now()
                        job.status = "completed"
                        job.cost = context.cost
                        job.save()
                    if isinstance(msg, Error):
                        raise Exception(msg.error)
                    yield msg.model_dump_json() + "\n"
                else:
                    await asyncio.sleep(0.1)

            while context.has_messages():
                msg = await context.pop_message_async()
                yield msg.model_dump_json() + "\n"

        except Exception as e:
            log.exception(e)
            job.finished_at = datetime.now()
            job.status = "failed"
            job.error = str(e)[:256]
            job.cost = context.cost
            job.save()
            yield json.dumps(
                JobUpdate(status="failed", error=str(e)).model_dump()
            ) + "\n"

    return StreamingResponse(generate(), media_type="application/x-ndjson")



"""Test the Processor class."""
import unittest
from PIL import Image

from controlnet_aux.processor import Processor


class TestProcessor(unittest.TestCase):
    def test_hed(self):
        processor = Processor('hed')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_midas(self):
        processor = Processor('midas')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_mlsd(self):
        processor = Processor('mlsd')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_openpose(self):
        processor = Processor('openpose')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_pidinet(self):
        processor = Processor('pidinet')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_normalbae(self):
        processor = Processor('normalbae')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_lineart(self):
        processor = Processor('lineart')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_lineart_coarse(self):
        processor = Processor('lineart_coarse')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_lineart_anime(self):
        processor = Processor('lineart_anime')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_canny(self):
        processor = Processor('canny')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_content_shuffle(self):
        processor = Processor('content_shuffle')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_zoe(self):
        processor = Processor('zoe')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_mediapipe_face(self):
        processor = Processor('mediapipe_face')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)

    def test_tile(self):
        processor = Processor('tile')
        image = Image.open('test_image.png')
        processed_image = processor(image)
        self.assertIsInstance(processed_image, bytes)


if __name__ == '__main__':
    unittest.main()


import os
import warnings

import cv2
import numpy as np
import torch
from einops import rearrange
from PIL import Image

from controlnet_aux.util import HWC3, nms, resize_image_with_pad, safe_step,common_input_validate, custom_hf_download, HF_MODEL_NAME
from .model import pidinet


class PidiNetDetector:
    def __init__(self, netNetwork):
        self.netNetwork = netNetwork

    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, filename="table5_pidinet.pth"):
        model_path = custom_hf_download(pretrained_model_or_path, filename)

        netNetwork = pidinet()
        netNetwork.load_state_dict({k.replace('module.', ''): v for k, v in torch.load(model_path)['state_dict'].items()})
        netNetwork.eval()

        return cls(netNetwork)

    def to(self, device):
        self.netNetwork.to(device)
        return self
    
    def __call__(self, input_image, detect_resolution=512, safe=False, output_type="pil", scribble=False, apply_filter=False, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        detected_map, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        device = next(iter(self.netNetwork.parameters())).device
        
        detected_map = detected_map[:, :, ::-1].copy()
        with torch.no_grad():
            image_pidi = torch.from_numpy(detected_map).float().to(device)
            image_pidi = image_pidi / 255.0
            image_pidi = rearrange(image_pidi, 'h w c -> 1 c h w')
            edge = self.netNetwork(image_pidi)[-1]
            edge = edge.cpu().numpy()
            if apply_filter:
                edge = edge > 0.5 
            if safe:
                edge = safe_step(edge)
            edge = (edge * 255.0).clip(0, 255).astype(np.uint8)

        detected_map = edge[0, 0]

        if scribble:
            detected_map = nms(detected_map, 127, 3.0)
            detected_map = cv2.GaussianBlur(detected_map, (0, 0), 3.0)
            detected_map[detected_map > 4] = 255
            detected_map[detected_map < 255] = 0

        detected_map = HWC3(remove_pad(detected_map))

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)

        return detected_map



import io
import librosa
from matplotlib import pyplot as plt
import numpy as np
from nodetool.metadata.types import Tensor
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.nodes.nodetool.audio.audio_helpers import (
    convert_to_float,
)
from nodetool.metadata.types import AudioRef
from nodetool.metadata.types import ImageRef
from nodetool.workflows.base_node import BaseNode
from pydantic import Field
from typing import Literal, Optional


class AmplitudeToDB(BaseNode):
    """
    Converts an amplitude spectrogram to a dB-scaled spectrogram.

    #### Applications
    - Audio Analysis: By converting to dB scale, it accentuates the variances in the data making it easier to analyze the audio signal.
    - Audio Classification: When classifying audio signals, converting to dB scale can help distinguish between different classes by amplifying subtle details.
    """

    tensor: Tensor = Field(
        description="The amplitude tensor to be converted to dB scale."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        db_tensor = librosa.amplitude_to_db(self.tensor.to_numpy(), ref=np.max)
        return Tensor.from_numpy(db_tensor)


class ChromaSTFT(BaseNode):
    """
    This node creates a chromagram from a waveform or power spectrogram to identify different pitch classes in an audio signal.

    #### Applications
    - Chord recognition in music: Helps in identifying the chords used in a piece of music.
    - Music genre classification: Assists in distinguishing the genre of a piece based on the unique pitch classes used.

    #### Notes
    - Parameters 'n_fft' and 'hop_length' control the resolution of the chromagram. A higher 'n_fft' or lower 'hop_length' increases frequency resolution, but decreases time resolution, and vice versa.
    """

    audio: AudioRef = Field(
        default=AudioRef(), description="The audio file to extract chromagram from."
    )
    n_fft: int = Field(
        default=2048, ge=0, description="The number of samples per frame."
    )
    hop_length: int = Field(
        default=512, ge=0, description="The number of samples between frames."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        (samples, sample_rate) = await context.audio_to_numpy(self.audio)
        samples = convert_to_float(samples)
        chromagram = librosa.feature.chroma_stft(
            y=samples, sr=sample_rate, n_fft=self.n_fft, hop_length=self.hop_length
        )
        return Tensor.from_numpy(chromagram)


class DBToAmplitude(BaseNode):
    """
    The DBToAmplitude node Converts a dB-scaled spectrogram to an amplitude spectrogram.

    #### Applications
    - Audio processing: Converting transformed audio data back to its original amplitude scale is crucial for further data processing or synthesis.
    - Noise reduction: If you've used a dB conversion for visualization or noise reduction purposes, this node can convert the data back to its original state.
    """

    tensor: Tensor = Field(
        description="The dB-scaled tensor to be converted to amplitude scale."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        amplitude_tensor = librosa.db_to_amplitude(self.tensor.to_numpy())
        return Tensor.from_numpy(amplitude_tensor)


class DBToPower(BaseNode):
    """
    This node converts a decibel (dB) spectrogram back to power scale.
    """

    tensor: Tensor = Field(
        default=Tensor(), description="The tensor containing the decibel spectrogram."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        db_spec = self.tensor.to_numpy()
        return Tensor.from_numpy(librosa.db_to_power(db_spec))


class GriffinLim(BaseNode):
    """
    GriffinLim Node performs phase reconstruction on a magnitude spectrogram utilizing the Griffin-Lim algorithm.

    #### Applications
    - Reconstructing audio signals from given spectrograms.
    - Restoring time-related elements in audio processing pipelines.
    """

    magnitude_spectrogram: Tensor = Field(
        description="Magnitude spectrogram input for phase reconstruction."
    )
    n_iter: int = Field(
        default=32, description="Number of iterations for the Griffin-Lim algorithm."
    )
    hop_length: int = Field(
        default=512, description="Number of samples between successive frames."
    )
    win_length: Optional[int] = Field(
        default=None,
        description="Each frame of audio is windowed by `window()`. The window will be of length `win_length` and then padded with zeros to match `n_fft`.",
    )
    window: str = Field(
        default="hann",
        description="Type of window to use for Griffin-Lim transformation.",
    )
    center: bool = Field(
        default=True,
        description="If True, the signal `y` is padded so that frame `D[:, t]` is centered at `y[t * hop_length]`.",
    )
    length: Optional[int] = Field(
        default=None,
        description="If given, the resulting signal will be zero-padded or clipped to this length.",
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        reconstructed_audio = librosa.griffinlim(
            S=self.magnitude_spectrogram.to_numpy(),
            n_iter=self.n_iter,
            hop_length=self.hop_length,
            win_length=self.win_length,
            window=self.window,
            center=self.center,
            length=self.length,
        )
        return Tensor.from_numpy(reconstructed_audio)


class MelSpectrogram(BaseNode):
    """
    MelSpecNode computes the Mel-frequency spectrogram for an audio signal.
    """

    audio: AudioRef = Field(
        default=AudioRef(), description="The audio file to convert to a tensor."
    )
    n_fft: int = Field(
        default=2048, ge=0, description="The number of samples per frame."
    )
    hop_length: int = Field(
        default=512, ge=0, description="The number of samples between frames."
    )
    n_mels: int = Field(
        default=128, ge=0, description="The number of Mel bands to generate."
    )
    fmin: int = Field(default=0, ge=0, description="The lowest frequency (in Hz).")
    fmax: int = Field(default=8000, ge=0, description="The highest frequency (in Hz).")

    async def process(self, context: ProcessingContext) -> Tensor:
        (samples, sample_rate) = await context.audio_to_numpy(self.audio)
        samples = convert_to_float(samples)
        melspectrogram = librosa.feature.melspectrogram(
            y=samples,
            sr=sample_rate,
            n_fft=self.n_fft,
            hop_length=self.hop_length,
            n_mels=self.n_mels,
            fmin=self.fmin,
            fmax=self.fmax,
        )
        return Tensor.from_numpy(melspectrogram)


class MFCC(BaseNode):
    """
    MFCC Node computes the Mel-frequency cepstral coefficients (MFCCs) from an audio signal.
    """

    audio: AudioRef = Field(
        default=AudioRef(), description="The audio file to extract MFCCs from."
    )
    n_mfcc: int = Field(default=13, ge=0, description="The number of MFCCs to extract.")
    n_fft: int = Field(
        default=2048, ge=0, description="The number of samples per frame."
    )
    hop_length: int = Field(
        default=512, ge=0, description="The number of samples between frames."
    )
    fmin: int = Field(default=0, ge=0, description="The lowest frequency (in Hz).")
    fmax: int = Field(default=8000, ge=0, description="The highest frequency (in Hz).")

    async def process(self, context: ProcessingContext) -> Tensor:
        (samples, sample_rate) = await context.audio_to_numpy(self.audio)
        samples = convert_to_float(samples)
        mfccs = librosa.feature.mfcc(
            y=samples,
            sr=sample_rate,
            n_mfcc=self.n_mfcc,
            n_fft=self.n_fft,
            hop_length=self.hop_length,
            fmin=self.fmin,
            fmax=self.fmax,
        )
        return Tensor.from_numpy(mfccs)


class PlotSpectrogram(BaseNode):
    """
    The PlotSpectrogram node generates a visual representation of the spectrum of frequencies in an audio signal as they vary with time.

    #### Applications
    - Audio Analysis: Allows users to visually see the spectrum of frequencies in their data.
    - Machine Learning: Used as a preprocessing step for feeding data into image-based ML models.
    - Sound engineering: Helps in identifying specific tones or frequencies in a music piece or a sound bite.
    """

    tensor: Tensor = Field(
        default=Tensor(), description="The tensor containing the mel spectrogram."
    )
    fmax: int = Field(default=8000, ge=0, description="The highest frequency (in Hz).")

    async def process(self, context: ProcessingContext) -> ImageRef:
        spec = self.tensor.to_numpy()
        plt.figure(figsize=(10, 10))
        librosa.display.specshow(spec, y_axis="mel", fmax=self.fmax, x_axis="time")
        plt.colorbar(format="%+2.0f dB")
        plt.title("Mel spectrogram")
        plt.tight_layout()
        buf = io.BytesIO()
        plt.savefig(buf, format="png")
        buf.seek(0)
        return await context.image_from_bytes(buf.getvalue())


class PowertToDB(BaseNode):
    """
    Converts a power spectrogram to decibel (dB) scale.

    #### Applications
    - Audio Signal Processing: The inputted power spectrogram is processed to convert it into a decibel scale which is utilised further.
    - Audio Visualization: The decibel-scaled spectrogram is a better fit for visualizations due to its reduced skewness.
    """

    tensor: Tensor = Field(
        default=Tensor(), description="The tensor containing the power spectrogram."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        power_spec = self.tensor.to_numpy()
        return Tensor.from_numpy(librosa.power_to_db(power_spec, ref=np.max))


class SpectralContrast(BaseNode):
    """
    The spectral contrast measures the difference in amplitude between the most noticeable parts (peaks) and the less noticeable parts (valleys) in a sound spectrum.

    #### Applications
    - Music genre classification: distinguishing between different types of music based on the color of sound.
    - Instrument recognition: recognizing different musical instruments by the difference in their spectral contrast.
    - Audio analysis: determining various characteristics of audio files.

    Useful note: The `n_fft` and `hop_length` parameters affect the resolution of the analysis. A higher `n_fft` provides better frequency resolution but worse time resolution, and vice versa for a lower `hop_length`.
    """

    audio: AudioRef = Field(
        default=AudioRef(),
        description="The audio file to extract spectral contrast from.",
    )
    n_fft: int = Field(
        default=2048, ge=0, description="The number of samples per frame."
    )
    hop_length: int = Field(
        default=512, ge=0, description="The number of samples between frames."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        (samples, sample_rate) = await context.audio_to_numpy(self.audio)
        samples = convert_to_float(samples)
        spectral_contrast = librosa.feature.spectral_contrast(
            y=samples, sr=sample_rate, n_fft=self.n_fft, hop_length=self.hop_length
        )
        return Tensor.from_numpy(spectral_contrast)


class STFT(BaseNode):
    """
    This node computes the Short-Time Fourier Transform (STFT) matrix for an audio signal.
    The STFT matrix represents the signal in both time and frequency domains, forming the foundation for many audio processing tasks.

    #### Applications
    - Audio Analysis: By transforming the audio signal into a visualizable format, it helps in understanding and analyzing the audio signal.
    - Sound Processing: It plays a key foundational role in sound effects, tuning, compression, and more.
    - Audio Feature Extraction: It can be used to analyze frequency-based features for sound classification.
    - Music Information Retrieval: It helps in music transcription, rhythm and tempo analysis.
    """

    audio: AudioRef = Field(
        default=AudioRef(),
        description="The audio file to compute the STFT matrix from.",
    )
    n_fft: int = Field(
        default=2048, ge=0, description="The number of samples per frame."
    )
    hop_length: int = Field(
        default=512, ge=0, description="The number of samples between frames."
    )
    win_length: Optional[int] = Field(
        default=None, description="The window length. If None, it defaults to n_fft."
    )
    window: str = Field(default="hann", description="The type of window to use.")
    center: bool = Field(
        default=True,
        description="If True, input signal is padded so that frame D[:, t] is centered at y[t * hop_length].",
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        (samples, sample_rate) = await context.audio_to_numpy(self.audio)
        samples = convert_to_float(samples)
        stft_matrix = librosa.stft(
            y=samples,
            n_fft=self.n_fft,
            hop_length=self.hop_length,
            win_length=self.win_length,
            window=self.window,
            center=self.center,
        )
        return Tensor.from_numpy(np.abs(stft_matrix))



import torch
import comfy.model_management
import comfy.conds

def prepare_mask(noise_mask, shape, device):
    """ensures noise mask is of proper dimensions"""
    noise_mask = torch.nn.functional.interpolate(noise_mask.reshape((-1, 1, noise_mask.shape[-2], noise_mask.shape[-1])), size=(shape[2], shape[3]), mode="bilinear")
    noise_mask = torch.cat([noise_mask] * shape[1], dim=1)
    noise_mask = comfy.utils.repeat_to_batch_size(noise_mask, shape[0])
    noise_mask = noise_mask.to(device)
    return noise_mask

def get_models_from_cond(cond, model_type):
    models = []
    for c in cond:
        if model_type in c:
            models += [c[model_type]]
    return models

def convert_cond(cond):
    out = []
    for c in cond:
        temp = c[1].copy()
        model_conds = temp.get("model_conds", {})
        if c[0] is not None:
            model_conds["c_crossattn"] = comfy.conds.CONDCrossAttn(c[0]) #TODO: remove
            temp["cross_attn"] = c[0]
        temp["model_conds"] = model_conds
        out.append(temp)
    return out

def get_additional_models(conds, dtype):
    """loads additional models in conditioning"""
    cnets = []
    gligen = []

    for k in conds:
        cnets += get_models_from_cond(conds[k], "control")
        gligen += get_models_from_cond(conds[k], "gligen")

    control_nets = set(cnets)

    inference_memory = 0
    control_models = []
    for m in control_nets:
        control_models += m.get_models()
        inference_memory += m.inference_memory_requirements(dtype)

    gligen = [x[1] for x in gligen]
    models = control_models + gligen
    return models, inference_memory

def cleanup_additional_models(models):
    """cleanup additional models that were loaded"""
    for m in models:
        if hasattr(m, 'cleanup'):
            m.cleanup()


def prepare_sampling(model, noise_shape, conds):
    device = model.load_device
    real_model = None
    models, inference_memory = get_additional_models(conds, model.model_dtype())
    comfy.model_management.load_models_gpu([model] + models, model.memory_required([noise_shape[0] * 2] + list(noise_shape[1:])) + inference_memory)
    real_model = model.model

    return real_model, conds, models

def cleanup_models(conds, models):
    cleanup_additional_models(models)

    control_cleanup = []
    for k in conds:
        control_cleanup += get_models_from_cond(conds[k], "control")

    cleanup_additional_models(set(control_cleanup))



from pydantic import Field
from nodetool.metadata.types import Sigmas
from nodetool.common.comfy_node import ComfyNode


class SplitSigmas(ComfyNode):
    sigmas: Sigmas = Field(
        default=Sigmas(), description="The array of sigmas to split."
    )
    step: int = Field(
        default=0, description="The specific step at which to split the sigmas array."
    )

    @classmethod
    def return_types(cls):
        return {"sigmas": Sigmas}, Sigmas


class FlipSigmas(ComfyNode):
    sigmas: Sigmas = Field(default=Sigmas(), description="The array of sigmas to flip.")

    @classmethod
    def return_type(cls):
        return {"sigmas": Sigmas}



from typing import Any
from nodetool.api.types.graph import Edge
from nodetool.workflows.base_node import BaseNode


def convert_to_comfy_workflow(edges: list[Edge], nodes: list[BaseNode]):
    """
    Converts a Nodetool workflow to a Comfy workflow.
    """
    json = {}
    for node in nodes:
        inputs = node.node_properties()
        inputs = {
            k: v.to_dict() if hasattr(v, "to_dict") else v
            for k, v in inputs.items()
            if v is not None
        }
        json[node._id] = {
            "class_type": node.__class__.__name__,
            "inputs": inputs,
        }

    def find_node(id):
        for node in nodes:
            if node._id == id:
                return node
        raise Exception(f"Could not find node with id {id}")

    def find_output(node: BaseNode, name: str):
        for index, output in enumerate(node.outputs()):
            if output.name == name:
                return index
        raise Exception(f"Could not find property {name} in node {node}")

    for edge in edges:
        source_node = find_node(edge.source)
        source_index = find_output(source_node, edge.sourceHandle)
        json[edge.target]["inputs"][edge.targetHandle] = [edge.source, source_index]

    return json



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class SaveAudio(GraphNode):
    value: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None, temp_id=None), description=None)
    folder: FolderRef | GraphNode | tuple[GraphNode, str] = Field(default=FolderRef(type='folder', uri='', asset_id=None, temp_id=None), description='The folder to save the audio file to. ')
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='audio', description='The name of the audio file.')
    @classmethod
    def get_node_type(cls): return "nodetool.audio.SaveAudio"





"""
This file contains the definition of the SMPL model

It is adapted from opensource project GraphCMR (https://github.com/nkolot/GraphCMR/)
"""
from __future__ import division

import torch
import torch.nn as nn
import numpy as np
import scipy.sparse
try:
    import cPickle as pickle
except ImportError:
    import pickle

from custom_mesh_graphormer.utils.geometric_layers import rodrigues
import custom_mesh_graphormer.modeling.data.config as cfg

from comfy.model_management import get_torch_device
from wrapper_for_mps import sparse_to_dense
device = get_torch_device()

class SMPL(nn.Module):

    def __init__(self, gender='neutral'):
        super(SMPL, self).__init__()

        if gender=='m':
            model_file=cfg.SMPL_Male
        elif gender=='f':
            model_file=cfg.SMPL_Female
        else:
            model_file=cfg.SMPL_FILE

        smpl_model = pickle.load(open(model_file, 'rb'), encoding='latin1') 
        J_regressor = smpl_model['J_regressor'].tocoo()
        row = J_regressor.row
        col = J_regressor.col
        data = J_regressor.data
        i = torch.LongTensor([row, col])
        v = torch.FloatTensor(data)
        J_regressor_shape = [24, 6890]
        self.register_buffer('J_regressor', torch.sparse_coo_tensor(i, v, J_regressor_shape).to_dense())
        self.register_buffer('weights', torch.FloatTensor(smpl_model['weights']))
        self.register_buffer('posedirs', torch.FloatTensor(smpl_model['posedirs']))
        self.register_buffer('v_template', torch.FloatTensor(smpl_model['v_template']))
        self.register_buffer('shapedirs', torch.FloatTensor(np.array(smpl_model['shapedirs'])))
        self.register_buffer('faces', torch.from_numpy(smpl_model['f'].astype(np.int64)))
        self.register_buffer('kintree_table', torch.from_numpy(smpl_model['kintree_table'].astype(np.int64)))
        id_to_col = {self.kintree_table[1, i].item(): i for i in range(self.kintree_table.shape[1])}
        self.register_buffer('parent', torch.LongTensor([id_to_col[self.kintree_table[0, it].item()] for it in range(1, self.kintree_table.shape[1])]))

        self.pose_shape = [24, 3]
        self.beta_shape = [10]
        self.translation_shape = [3]

        self.pose = torch.zeros(self.pose_shape)
        self.beta = torch.zeros(self.beta_shape)
        self.translation = torch.zeros(self.translation_shape)

        self.verts = None
        self.J = None
        self.R = None
        
        J_regressor_extra = torch.from_numpy(np.load(cfg.JOINT_REGRESSOR_TRAIN_EXTRA)).float()
        self.register_buffer('J_regressor_extra', J_regressor_extra)
        self.joints_idx = cfg.JOINTS_IDX

        J_regressor_h36m_correct = torch.from_numpy(np.load(cfg.JOINT_REGRESSOR_H36M_correct)).float()
        self.register_buffer('J_regressor_h36m_correct', J_regressor_h36m_correct)


    def forward(self, pose, beta):
        device = pose.device
        batch_size = pose.shape[0]
        v_template = self.v_template[None, :]
        shapedirs = self.shapedirs.view(-1,10)[None, :].expand(batch_size, -1, -1)
        beta = beta[:, :, None]
        v_shaped = torch.matmul(shapedirs, beta).view(-1, 6890, 3) + v_template
        # batched sparse matmul not supported in pytorch
        J = []
        for i in range(batch_size):
            J.append(torch.matmul(self.J_regressor, v_shaped[i]))
        J = torch.stack(J, dim=0)
        # input it rotmat: (bs,24,3,3)
        if pose.ndimension() == 4:
            R = pose
        # input it rotmat: (bs,72)
        elif pose.ndimension() == 2:
            pose_cube = pose.view(-1, 3) # (batch_size * 24, 1, 3)
            R = rodrigues(pose_cube).view(batch_size, 24, 3, 3)
            R = R.view(batch_size, 24, 3, 3)
        I_cube = torch.eye(3)[None, None, :].to(device)
        # I_cube = torch.eye(3)[None, None, :].expand(theta.shape[0], R.shape[1]-1, -1, -1)
        lrotmin = (R[:,1:,:] - I_cube).view(batch_size, -1)
        posedirs = self.posedirs.view(-1,207)[None, :].expand(batch_size, -1, -1)
        v_posed = v_shaped + torch.matmul(posedirs, lrotmin[:, :, None]).view(-1, 6890, 3)
        J_ = J.clone()
        J_[:, 1:, :] = J[:, 1:, :] - J[:, self.parent, :]
        G_ = torch.cat([R, J_[:, :, :, None]], dim=-1)
        pad_row = torch.FloatTensor([0,0,0,1]).to(device).view(1,1,1,4).expand(batch_size, 24, -1, -1)
        G_ = torch.cat([G_, pad_row], dim=2)
        G = [G_[:, 0].clone()]
        for i in range(1, 24):
            G.append(torch.matmul(G[self.parent[i-1]], G_[:, i, :, :]))
        G = torch.stack(G, dim=1)

        rest = torch.cat([J, torch.zeros(batch_size, 24, 1).to(device)], dim=2).view(batch_size, 24, 4, 1)
        zeros = torch.zeros(batch_size, 24, 4, 3).to(device)
        rest = torch.cat([zeros, rest], dim=-1)
        rest = torch.matmul(G, rest)
        G = G - rest
        T = torch.matmul(self.weights, G.permute(1,0,2,3).contiguous().view(24,-1)).view(6890, batch_size, 4, 4).transpose(0,1)
        rest_shape_h = torch.cat([v_posed, torch.ones_like(v_posed)[:, :, [0]]], dim=-1)
        v = torch.matmul(T, rest_shape_h[:, :, :, None])[:, :, :3, 0]
        return v

    def get_joints(self, vertices):
        """
        This method is used to get the joint locations from the SMPL mesh
        Input:
            vertices: size = (B, 6890, 3)
        Output:
            3D joints: size = (B, 38, 3)
        """
        joints = torch.einsum('bik,ji->bjk', [vertices, self.J_regressor])
        joints_extra = torch.einsum('bik,ji->bjk', [vertices, self.J_regressor_extra])
        joints = torch.cat((joints, joints_extra), dim=1)
        joints = joints[:, cfg.JOINTS_IDX]
        return joints

    def get_h36m_joints(self, vertices):
        """
        This method is used to get the joint locations from the SMPL mesh
        Input:
            vertices: size = (B, 6890, 3)
        Output:
            3D joints: size = (B, 24, 3)
        """
        joints = torch.einsum('bik,ji->bjk', [vertices, self.J_regressor_h36m_correct])
        return joints

class SparseMM(torch.autograd.Function):
    """Redefine sparse @ dense matrix multiplication to enable backpropagation.
    The builtin matrix multiplication operation does not support backpropagation in some cases.
    """
    @staticmethod
    def forward(ctx, sparse, dense):
        ctx.req_grad = dense.requires_grad
        ctx.save_for_backward(sparse)
        return torch.matmul(sparse, dense)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = None
        sparse, = ctx.saved_tensors
        if ctx.req_grad:
            grad_input = torch.matmul(sparse.t(), grad_output)
        return None, grad_input

def spmm(sparse, dense):
    sparse = sparse.to(device)
    dense = dense.to(device)
    return SparseMM.apply(sparse, dense)


def scipy_to_pytorch(A, U, D):
    """Convert scipy sparse matrices to pytorch sparse matrix."""
    ptU = []
    ptD = []
    
    for i in range(len(U)):
        u = scipy.sparse.coo_matrix(U[i])
        i = torch.LongTensor(np.array([u.row, u.col]))
        v = torch.FloatTensor(u.data)
        ptU.append(sparse_to_dense(torch.sparse_coo_tensor(i, v, u.shape)))
    
    for i in range(len(D)):
        d = scipy.sparse.coo_matrix(D[i])
        i = torch.LongTensor(np.array([d.row, d.col]))
        v = torch.FloatTensor(d.data)
        ptD.append(sparse_to_dense(torch.sparse_coo_tensor(i, v, d.shape))) 

    return ptU, ptD


def adjmat_sparse(adjmat, nsize=1):
    """Create row-normalized sparse graph adjacency matrix."""
    adjmat = scipy.sparse.csr_matrix(adjmat)
    if nsize > 1:
        orig_adjmat = adjmat.copy()
        for _ in range(1, nsize):
            adjmat = adjmat * orig_adjmat
    adjmat.data = np.ones_like(adjmat.data)
    for i in range(adjmat.shape[0]):
        adjmat[i,i] = 1
    num_neighbors = np.array(1 / adjmat.sum(axis=-1))
    adjmat = adjmat.multiply(num_neighbors)
    adjmat = scipy.sparse.coo_matrix(adjmat)
    row = adjmat.row
    col = adjmat.col
    data = adjmat.data
    i = torch.LongTensor(np.array([row, col]))
    v = torch.from_numpy(data).float()
    adjmat = sparse_to_dense(torch.sparse_coo_tensor(i, v, adjmat.shape))
    return adjmat

def get_graph_params(filename, nsize=1):
    """Load and process graph adjacency matrix and upsampling/downsampling matrices."""
    data = np.load(filename, encoding='latin1', allow_pickle=True)
    A = data['A']
    U = data['U']
    D = data['D']
    U, D = scipy_to_pytorch(A, U, D)
    A = [adjmat_sparse(a, nsize=nsize) for a in A]
    return A, U, D


class Mesh(object):
    """Mesh object that is used for handling certain graph operations."""
    def __init__(self, filename=cfg.SMPL_sampling_matrix,
                 num_downsampling=1, nsize=1, device=torch.device('cuda')):
        self._A, self._U, self._D = get_graph_params(filename=filename, nsize=nsize)
        # self._A = [a.to(device) for a in self._A]
        self._U = [u.to(device) for u in self._U]
        self._D = [d.to(device) for d in self._D]
        self.num_downsampling = num_downsampling

        # load template vertices from SMPL and normalize them
        smpl = SMPL()
        ref_vertices = smpl.v_template
        center = 0.5*(ref_vertices.max(dim=0)[0] + ref_vertices.min(dim=0)[0])[None]
        ref_vertices -= center
        ref_vertices /= ref_vertices.abs().max().item()

        self._ref_vertices = ref_vertices.to(device)
        self.faces = smpl.faces.int().to(device)

    # @property
    # def adjmat(self):
    #     """Return the graph adjacency matrix at the specified subsampling level."""
    #     return self._A[self.num_downsampling].float()

    @property
    def ref_vertices(self):
        """Return the template vertices at the specified subsampling level."""
        ref_vertices = self._ref_vertices
        for i in range(self.num_downsampling):
            ref_vertices = torch.spmm(self._D[i], ref_vertices)
        return ref_vertices

    def downsample(self, x, n1=0, n2=None):
        """Downsample mesh."""
        if n2 is None:
            n2 = self.num_downsampling
        if x.ndimension() < 3:
            for i in range(n1, n2):
                x = spmm(self._D[i], x)
        elif x.ndimension() == 3:
            out = []
            for i in range(x.shape[0]):
                y = x[i]
                for j in range(n1, n2):
                    y = spmm(self._D[j], y)
                out.append(y)
            x = torch.stack(out, dim=0)
        return x

    def upsample(self, x, n1=1, n2=0):
        """Upsample mesh."""
        if x.ndimension() < 3:
            for i in reversed(range(n2, n1)):
                x = spmm(self._U[i], x)
        elif x.ndimension() == 3:
            out = []
            for i in range(x.shape[0]):
                y = x[i]
                for j in reversed(range(n2, n1)):
                    y = spmm(self._U[j], y)
                out.append(y)
            x = torch.stack(out, dim=0)
        return x



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.comfy.ipadapter import WeightTypeEnum

class IPAdapterApply(GraphNode):
    ipadapter: IPAdapter | GraphNode | tuple[GraphNode, str] = Field(default=IPAdapter(type='comfy.ip_adapter'), description='The IPAdapter to apply.')
    clip_vision: CLIPVision | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVision(type='comfy.clip_vision'), description='The CLIP vision to use.')
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to use.')
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to apply the IPAdapter to.')
    weight: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight of the application.')
    noise: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The amount of noise to apply.')
    weight_type: WeightTypeEnum | GraphNode | tuple[GraphNode, str] = Field(default=WeightTypeEnum('original'), description='The type of weight to apply.')
    start_at: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The starting point for applying the IPAdapter.')
    end_at: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The ending point for applying the IPAdapter.')
    unfold_batch: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Whether to unfold the batch during the application.')
    attn_mask: nodetool.metadata.types.Mask | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The optional attention mask to use (if any).')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterApply"


from nodetool.nodes.comfy.ipadapter import WeightTypeEnum

class IPAdapterApplyEncoded(GraphNode):
    ipadapter: IPAdapter | GraphNode | tuple[GraphNode, str] = Field(default=IPAdapter(type='comfy.ip_adapter'), description='The IPAdapter to apply.')
    clip_vision: CLIPVision | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVision(type='comfy.clip_vision'), description='The CLIP vision to use.')
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to use.')
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to apply the IPAdapter to.')
    weight: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight of the application.')
    noise: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The amount of noise to apply.')
    weight_type: WeightTypeEnum | GraphNode | tuple[GraphNode, str] = Field(default=WeightTypeEnum('original'), description='The type of weight to apply.')
    start_at: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The starting point for applying the IPAdapter.')
    end_at: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The ending point for applying the IPAdapter.')
    unfold_batch: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Whether to unfold the batch during the application.')
    attn_mask: nodetool.metadata.types.Mask | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The optional attention mask to use (if any).')
    embeds: Embeds | GraphNode | tuple[GraphNode, str] = Field(default=Embeds(type='comfy.embeds'), description='The encoded embeddings to apply.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterApplyEncoded"


from nodetool.nodes.comfy.ipadapter import WeightTypeEnum

class IPAdapterApplyFaceID(GraphNode):
    ipadapter: IPAdapter | GraphNode | tuple[GraphNode, str] = Field(default=IPAdapter(type='comfy.ip_adapter'), description='The IPAdapter.')
    clip_vision: CLIPVision | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVision(type='comfy.clip_vision'), description='The CLIP vision.')
    insightface: InsightFace | GraphNode | tuple[GraphNode, str] = Field(default=InsightFace(type='comfy.insight_face'), description='The InsightFace.')
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to use.')
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to apply the IPAdapter.')
    weight: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight for processing.')
    noise: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The noise level for processing.')
    weight_type: WeightTypeEnum | GraphNode | tuple[GraphNode, str] = Field(default=WeightTypeEnum('original'), description='The weight type to use.')
    start_at: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='Start applying from this step (normalized).')
    end_at: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='Ends applying at this step (normalized).')
    faceid_v2: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Flag to use faceId v2 algorithm.')
    weight_v2: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight for processing with faceId v2.')
    unfold_batch: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Flag to unfold batches for processing.')
    attn_mask: nodetool.metadata.types.Mask | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The attention mask.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterApplyFaceID"



class IPAdapterBatchEmbeds(GraphNode):
    embed1: Embeds | GraphNode | tuple[GraphNode, str] = Field(default=Embeds(type='comfy.embeds'), description='The first set of embeddings.')
    embed2: Embeds | GraphNode | tuple[GraphNode, str] = Field(default=Embeds(type='comfy.embeds'), description='The second set of embeddings.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterBatchEmbeds"



class IPAdapterEncoder(GraphNode):
    clip_vision: CLIPVision | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVision(type='comfy.clip_vision'), description='The CLIP vision to use.')
    image_1: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The first image to encode.')
    ipadapter_plus: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Whether to use IPAdapter+ enhancements.')
    noise: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The amount of noise to apply.')
    weight_1: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight for the first image.')
    image_2: nodetool.metadata.types.ImageTensor | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The second image to encode (optional).')
    image_3: nodetool.metadata.types.ImageTensor | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The third image to encode (optional).')
    image_4: nodetool.metadata.types.ImageTensor | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The fourth image to encode (optional).')
    weight_2: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight for the second image (optional).')
    weight_3: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight for the third image (optional).')
    weight_4: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The weight for the fourth image (optional).')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterEncoder"



class IPAdapterModelLoader(GraphNode):
    ipadapter_file: IPAdapterFile | GraphNode | tuple[GraphNode, str] = Field(default=IPAdapterFile(type='comfy.ip_adapter_file', name=''), description='List of available IPAdapter model names.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterModelLoader"



class IPAdapterSaveEmbeds(GraphNode):
    embeds: Embeds | GraphNode | tuple[GraphNode, str] = Field(default=Embeds(type='comfy.embeds'), description='The embeddings to save.')
    filename_prefix: str | GraphNode | tuple[GraphNode, str] = Field(default='embeds/IPAdapter', description='The prefix for the filename to save the embeddings.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.IPAdapterSaveEmbeds"


from nodetool.nodes.comfy.ipadapter import ProviderEnum

class InsightFaceLoader(GraphNode):
    provider: ProviderEnum | GraphNode | tuple[GraphNode, str] = Field(default=ProviderEnum('CPU'), description=None)
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.InsightFaceLoader"


from nodetool.nodes.comfy.ipadapter import InterpolationMethod
from nodetool.nodes.comfy.ipadapter import CropPosition

class PrepImageForClipVision(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to prepare.')
    interpolation: InterpolationMethod | GraphNode | tuple[GraphNode, str] = Field(default=InterpolationMethod('LANCZOS'), description='The interpolation method to use.')
    crop_position: CropPosition | GraphNode | tuple[GraphNode, str] = Field(default=CropPosition('center'), description='The crop position to use.')
    sharpening: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The amount of sharpening to apply.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.PrepImageForClipVision"


from nodetool.nodes.comfy.ipadapter import CropPosition

class PrepImageForInsightFace(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to prepare.')
    crop_position: CropPosition | GraphNode | tuple[GraphNode, str] = Field(default=CropPosition('center'), description='The crop position to use.')
    sharpening: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The amount of sharpening to apply.')
    pad_around: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Whether to pad around the image.')
    @classmethod
    def get_node_type(cls): return "comfy.ipadapter.PrepImageForInsightFace"





import os
import wandb
from custom_detectron2.utils import comm
from custom_detectron2.utils.events import EventWriter, get_event_storage


def setup_wandb(cfg, args):
    if comm.is_main_process():
        init_args = {
            k.lower(): v
            for k, v in cfg.WANDB.items()
            if isinstance(k, str) and k not in ["config"]
        }
        # only include most related part to avoid too big table
        # TODO: add configurable params to select which part of `cfg` should be saved in config
        if "config_exclude_keys" in init_args:
            init_args["config"] = cfg
            init_args["config"]["cfg_file"] = args.config_file
        else:
            init_args["config"] = {
                "model": cfg.MODEL,
                "solver": cfg.SOLVER,
                "cfg_file": args.config_file,
            }
        if ("name" not in init_args) or (init_args["name"] is None):
            init_args["name"] = os.path.basename(args.config_file)
        else:
            init_args["name"] = init_args["name"] + '_' + os.path.basename(args.config_file)
        wandb.init(**init_args)


class BaseRule(object):
    def __call__(self, target):
        return target


class IsIn(BaseRule):
    def __init__(self, keyword: str):
        self.keyword = keyword

    def __call__(self, target):
        return self.keyword in target


class Prefix(BaseRule):
    def __init__(self, keyword: str):
        self.keyword = keyword

    def __call__(self, target):
        return "/".join([self.keyword, target])


class WandbWriter(EventWriter):
    """
    Write all scalars to a tensorboard file.
    """

    def __init__(self):
        """
        Args:
            log_dir (str): the directory to save the output events
            kwargs: other arguments passed to `torch.utils.tensorboard.SummaryWriter(...)`
        """
        self._last_write = -1
        self._group_rules = [
            (IsIn("/"), BaseRule()),
            (IsIn("loss"), Prefix("train")),
        ]

    def write(self):

        storage = get_event_storage()

        def _group_name(scalar_name):
            for (rule, op) in self._group_rules:
                if rule(scalar_name):
                    return op(scalar_name)
            return scalar_name

        stats = {
            _group_name(name): scalars[0]
            for name, scalars in storage.latest().items()
            if scalars[1] > self._last_write
        }
        if len(stats) > 0:
            self._last_write = max([v[1] for k, v in storage.latest().items()])

        # storage.put_{image,histogram} is only meant to be used by
        # tensorboard writer. So we access its internal fields directly from here.
        if len(storage._vis_data) >= 1:
            stats["image"] = [
                wandb.Image(img, caption=img_name)
                for img_name, img, step_num in storage._vis_data
            ]
            # Storage stores all image data and rely on this writer to clear them.
            # As a result it assumes only one writer will use its image data.
            # An alternative design is to let storage store limited recent
            # data (e.g. only the most recent image) that all writers can access.
            # In that case a writer may not see all image data if its period is long.
            storage.clear_images()

        if len(storage._histograms) >= 1:

            def create_bar(tag, bucket_limits, bucket_counts, **kwargs):
                data = [
                    [label, val] for (label, val) in zip(bucket_limits, bucket_counts)
                ]
                table = wandb.Table(data=data, columns=["label", "value"])
                return wandb.plot.bar(table, "label", "value", title=tag)

            stats["hist"] = [create_bar(**params) for params in storage._histograms]

            storage.clear_histograms()

        if len(stats) == 0:
            return
        wandb.log(stats, step=storage.iter)

    def close(self):
        wandb.finish()


import json
from typing import Any, Literal
import pandas as pd
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import DataframeRef
from nodetool.workflows.base_node import BaseNode


class Access(BaseNode):
    """
    Retrieves a specific value from a dictionary using a key.
    dictionary, dict, select, get, value, access, key
    """

    _layout = "small"
    _primary_field = "key"

    dictionary: dict[(str, Any)] = {}
    key: str = ""

    async def process(self, context: ProcessingContext) -> Any:
        return self.dictionary[self.key]


class Update(BaseNode):
    """
    Updates a dictionary with one or more new key-value pairs.
    dictionary, dict, add, insert
    """

    _layout = "small"
    _primary_field = "new_pairs"

    dictionary: dict[(str, Any)] = {}
    new_pairs: dict[(str, Any)] = {}

    async def process(self, context: ProcessingContext) -> dict[(str, Any)]:
        self.dictionary.update(self.new_pairs)
        return self.dictionary


class Delete(BaseNode):
    """
    Removes a key-value pair from a dictionary based on the specified key.
    dictionary, dict, remove
    """

    _layout = "small"
    _primary_field = "key"

    dictionary: dict[(str, Any)] = {}
    key: str = ""

    async def process(self, context: ProcessingContext) -> dict[(str, Any)]:
        del self.dictionary[self.key]
        return self.dictionary


class DictFromJson(BaseNode):
    """
    Converts a JSON string into a dictionary.
    dictionary, json, parse
    """

    _layout = "small"

    json_string: str = ""

    async def process(self, context: ProcessingContext) -> dict[(str, Any)]:
        res = json.loads(self.json_string)
        if not isinstance(res, dict):
            raise ValueError("Input JSON is not a dictionary")
        return res


class DictFromList(BaseNode):
    """
    Generates a dictionary by pairing lists of keys and values.
    dictionary, create, keys, values
    """

    _layout = "small"

    keys: list[Any] = []
    values: list[Any] = []

    async def process(self, context: ProcessingContext) -> dict[Any, Any]:
        return dict(zip(self.keys, self.values))


class Merge(BaseNode):
    """
    Combines two dictionaries into one. Note: Values from the second input override duplicates.
    dictionary, merge, combine
    """

    _layout = "small"

    dict_a: dict[(str, Any)] = {}
    dict_b: dict[(str, Any)] = {}

    async def process(self, context: ProcessingContext) -> dict[(str, Any)]:
        return {**self.dict_a, **self.dict_b}


class SelectKeys(BaseNode):
    """
    Filters a dictionary to include only specified keys.
    dictionary, keys, filter
    """

    dictionary: dict[(str, Any)] = {}
    keys: list[str] = []

    async def process(self, context: ProcessingContext) -> dict[(str, Any)]:
        return {key: self.dictionary[key] for key in self.keys}


class DictToDataframe(BaseNode):
    """
    Converts a dictionary into a dataframe, each value in the dictionary becomes a column.
    dictionary, dataframe, pandas
    """

    _layout = "small"

    dictionary: dict[(str, Any)] = {}

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = pd.DataFrame([self.model_dump()])
        return await context.dataframe_from_pandas(df)


class RowsToDataframe(BaseNode):
    """
    Converts a list of dictionaries into a dataframe, each dictionary becomes a row.
    dictionary, dataframe, pandas
    """

    _layout = "small"

    rows: list[dict[(str, Any)]] = []

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = pd.DataFrame(self.rows)
        return await context.dataframe_from_pandas(df)



import math
from typing import List, NamedTuple, Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy.ndimage.filters import gaussian_filter

from . import util
from .model import bodypose_model


class Keypoint(NamedTuple):
    x: float
    y: float
    score: float = 1.0
    id: int = -1


class BodyResult(NamedTuple):
    # Note: Using `Union` instead of `|` operator as the ladder is a Python
    # 3.10 feature.
    # Annotator code should be Python 3.8 Compatible, as controlnet repo uses
    # Python 3.8 environment.
    # https://github.com/lllyasviel/ControlNet/blob/d3284fcd0972c510635a4f5abe2eeb71dc0de524/environment.yaml#L6
    keypoints: List[Union[Keypoint, None]]
    total_score: float
    total_parts: int


class Body(object):
    def __init__(self, model_path):
        self.model = bodypose_model()
        model_dict = util.transfer(self.model, torch.load(model_path))
        self.model.load_state_dict(model_dict)
        self.model.eval()

    def to(self, device):
        self.model.to(device)
        return self

    def __call__(self, oriImg):
        device = next(iter(self.model.parameters())).device
        # scale_search = [0.5, 1.0, 1.5, 2.0]
        scale_search = [0.5]
        boxsize = 368
        stride = 8
        padValue = 128
        thre1 = 0.1
        thre2 = 0.05
        multiplier = [x * boxsize / oriImg.shape[0] for x in scale_search]
        heatmap_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 19))
        paf_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 38))

        for m in range(len(multiplier)):
            scale = multiplier[m]
            imageToTest = util.smart_resize_k(oriImg, fx=scale, fy=scale)
            imageToTest_padded, pad = util.padRightDownCorner(imageToTest, stride, padValue)
            im = np.transpose(np.float32(imageToTest_padded[:, :, :, np.newaxis]), (3, 2, 0, 1)) / 256 - 0.5
            im = np.ascontiguousarray(im)

            data = torch.from_numpy(im).float()
            data = data.to(device)
            # data = data.permute([2, 0, 1]).unsqueeze(0).float()
            with torch.no_grad():
                Mconv7_stage6_L1, Mconv7_stage6_L2 = self.model(data)
            Mconv7_stage6_L1 = Mconv7_stage6_L1.cpu().numpy()
            Mconv7_stage6_L2 = Mconv7_stage6_L2.cpu().numpy()

            # extract outputs, resize, and remove padding
            # heatmap = np.transpose(np.squeeze(net.blobs[output_blobs.keys()[1]].data), (1, 2, 0))  # output 1 is heatmaps
            heatmap = np.transpose(np.squeeze(Mconv7_stage6_L2), (1, 2, 0))  # output 1 is heatmaps
            heatmap = util.smart_resize_k(heatmap, fx=stride, fy=stride)
            heatmap = heatmap[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
            heatmap = util.smart_resize(heatmap, (oriImg.shape[0], oriImg.shape[1]))

            # paf = np.transpose(np.squeeze(net.blobs[output_blobs.keys()[0]].data), (1, 2, 0))  # output 0 is PAFs
            paf = np.transpose(np.squeeze(Mconv7_stage6_L1), (1, 2, 0))  # output 0 is PAFs
            paf = util.smart_resize_k(paf, fx=stride, fy=stride)
            paf = paf[:imageToTest_padded.shape[0] - pad[2], :imageToTest_padded.shape[1] - pad[3], :]
            paf = util.smart_resize(paf, (oriImg.shape[0], oriImg.shape[1]))

            heatmap_avg += heatmap_avg + heatmap / len(multiplier)
            paf_avg += + paf / len(multiplier)

        all_peaks = []
        peak_counter = 0

        for part in range(18):
            map_ori = heatmap_avg[:, :, part]
            one_heatmap = gaussian_filter(map_ori, sigma=3)

            map_left = np.zeros(one_heatmap.shape)
            map_left[1:, :] = one_heatmap[:-1, :]
            map_right = np.zeros(one_heatmap.shape)
            map_right[:-1, :] = one_heatmap[1:, :]
            map_up = np.zeros(one_heatmap.shape)
            map_up[:, 1:] = one_heatmap[:, :-1]
            map_down = np.zeros(one_heatmap.shape)
            map_down[:, :-1] = one_heatmap[:, 1:]

            peaks_binary = np.logical_and.reduce(
                (one_heatmap >= map_left, one_heatmap >= map_right, one_heatmap >= map_up, one_heatmap >= map_down, one_heatmap > thre1))
            peaks = list(zip(np.nonzero(peaks_binary)[1], np.nonzero(peaks_binary)[0]))  # note reverse
            peaks_with_score = [x + (map_ori[x[1], x[0]],) for x in peaks]
            peak_id = range(peak_counter, peak_counter + len(peaks))
            peaks_with_score_and_id = [peaks_with_score[i] + (peak_id[i],) for i in range(len(peak_id))]

            all_peaks.append(peaks_with_score_and_id)
            peak_counter += len(peaks)

        # find connection in the specified sequence, center 29 is in the position 15
        limbSeq = [[2, 3], [2, 6], [3, 4], [4, 5], [6, 7], [7, 8], [2, 9], [9, 10], \
                   [10, 11], [2, 12], [12, 13], [13, 14], [2, 1], [1, 15], [15, 17], \
                   [1, 16], [16, 18], [3, 17], [6, 18]]
        # the middle joints heatmap correpondence
        mapIdx = [[31, 32], [39, 40], [33, 34], [35, 36], [41, 42], [43, 44], [19, 20], [21, 22], \
                  [23, 24], [25, 26], [27, 28], [29, 30], [47, 48], [49, 50], [53, 54], [51, 52], \
                  [55, 56], [37, 38], [45, 46]]

        connection_all = []
        special_k = []
        mid_num = 10

        for k in range(len(mapIdx)):
            score_mid = paf_avg[:, :, [x - 19 for x in mapIdx[k]]]
            candA = all_peaks[limbSeq[k][0] - 1]
            candB = all_peaks[limbSeq[k][1] - 1]
            nA = len(candA)
            nB = len(candB)
            indexA, indexB = limbSeq[k]
            if (nA != 0 and nB != 0):
                connection_candidate = []
                for i in range(nA):
                    for j in range(nB):
                        vec = np.subtract(candB[j][:2], candA[i][:2])
                        norm = math.sqrt(vec[0] * vec[0] + vec[1] * vec[1])
                        norm = max(0.001, norm)
                        vec = np.divide(vec, norm)

                        startend = list(zip(np.linspace(candA[i][0], candB[j][0], num=mid_num), \
                                            np.linspace(candA[i][1], candB[j][1], num=mid_num)))

                        vec_x = np.array([score_mid[int(round(startend[I][1])), int(round(startend[I][0])), 0] \
                                          for I in range(len(startend))])
                        vec_y = np.array([score_mid[int(round(startend[I][1])), int(round(startend[I][0])), 1] \
                                          for I in range(len(startend))])

                        score_midpts = np.multiply(vec_x, vec[0]) + np.multiply(vec_y, vec[1])
                        score_with_dist_prior = sum(score_midpts) / len(score_midpts) + min(
                            0.5 * oriImg.shape[0] / norm - 1, 0)
                        criterion1 = len(np.nonzero(score_midpts > thre2)[0]) > 0.8 * len(score_midpts)
                        criterion2 = score_with_dist_prior > 0
                        if criterion1 and criterion2:
                            connection_candidate.append(
                                [i, j, score_with_dist_prior, score_with_dist_prior + candA[i][2] + candB[j][2]])

                connection_candidate = sorted(connection_candidate, key=lambda x: x[2], reverse=True)
                connection = np.zeros((0, 5))
                for c in range(len(connection_candidate)):
                    i, j, s = connection_candidate[c][0:3]
                    if (i not in connection[:, 3] and j not in connection[:, 4]):
                        connection = np.vstack([connection, [candA[i][3], candB[j][3], s, i, j]])
                        if (len(connection) >= min(nA, nB)):
                            break

                connection_all.append(connection)
            else:
                special_k.append(k)
                connection_all.append([])

        # last number in each row is the total parts number of that person
        # the second last number in each row is the score of the overall configuration
        subset = -1 * np.ones((0, 20))
        candidate = np.array([item for sublist in all_peaks for item in sublist])

        for k in range(len(mapIdx)):
            if k not in special_k:
                partAs = connection_all[k][:, 0]
                partBs = connection_all[k][:, 1]
                indexA, indexB = np.array(limbSeq[k]) - 1

                for i in range(len(connection_all[k])):  # = 1:size(temp,1)
                    found = 0
                    subset_idx = [-1, -1]
                    for j in range(len(subset)):  # 1:size(subset,1):
                        if subset[j][indexA] == partAs[i] or subset[j][indexB] == partBs[i]:
                            subset_idx[found] = j
                            found += 1

                    if found == 1:
                        j = subset_idx[0]
                        if subset[j][indexB] != partBs[i]:
                            subset[j][indexB] = partBs[i]
                            subset[j][-1] += 1
                            subset[j][-2] += candidate[partBs[i].astype(int), 2] + connection_all[k][i][2]
                    elif found == 2:  # if found 2 and disjoint, merge them
                        j1, j2 = subset_idx
                        membership = ((subset[j1] >= 0).astype(int) + (subset[j2] >= 0).astype(int))[:-2]
                        if len(np.nonzero(membership == 2)[0]) == 0:  # merge
                            subset[j1][:-2] += (subset[j2][:-2] + 1)
                            subset[j1][-2:] += subset[j2][-2:]
                            subset[j1][-2] += connection_all[k][i][2]
                            subset = np.delete(subset, j2, 0)
                        else:  # as like found == 1
                            subset[j1][indexB] = partBs[i]
                            subset[j1][-1] += 1
                            subset[j1][-2] += candidate[partBs[i].astype(int), 2] + connection_all[k][i][2]

                    # if find no partA in the subset, create a new subset
                    elif not found and k < 17:
                        row = -1 * np.ones(20)
                        row[indexA] = partAs[i]
                        row[indexB] = partBs[i]
                        row[-1] = 2
                        row[-2] = sum(candidate[connection_all[k][i, :2].astype(int), 2]) + connection_all[k][i][2]
                        subset = np.vstack([subset, row])
        # delete some rows of subset which has few parts occur
        deleteIdx = []
        for i in range(len(subset)):
            if subset[i][-1] < 4 or subset[i][-2] / subset[i][-1] < 0.4:
                deleteIdx.append(i)
        subset = np.delete(subset, deleteIdx, axis=0)

        # subset: n*20 array, 0-17 is the index in candidate, 18 is the total score, 19 is the total parts
        # candidate: x, y, score, id
        return candidate, subset
    
    @staticmethod
    def format_body_result(candidate: np.ndarray, subset: np.ndarray) -> List[BodyResult]:
        """
        Format the body results from the candidate and subset arrays into a list of BodyResult objects.
        
        Args:
            candidate (np.ndarray): An array of candidates containing the x, y coordinates, score, and id
                for each body part.
            subset (np.ndarray): An array of subsets containing indices to the candidate array for each
                person detected. The last two columns of each row hold the total score and total parts
                of the person.

        Returns:
            List[BodyResult]: A list of BodyResult objects, where each object represents a person with
                detected keypoints, total score, and total parts.
        """
        return [
            BodyResult(
                keypoints=[
                    Keypoint(
                        x=candidate[candidate_index][0],
                        y=candidate[candidate_index][1],
                        score=candidate[candidate_index][2],
                        id=candidate[candidate_index][3]
                    ) if candidate_index != -1 else None
                    for candidate_index in person[:18].astype(int)
                ],
                total_score=person[18],
                total_parts=person[19]
            )
            for person in subset
        ]
    

if __name__ == "__main__":
    body_estimation = Body('../model/body_pose_model.pth')

    test_image = '../images/ski.jpg'
    oriImg = cv2.imread(test_image)  # B,G,R order
    candidate, subset = body_estimation(oriImg)
    bodies = body_estimation.format_body_result(candidate, subset)

    canvas = oriImg
    for body in bodies:
        canvas = util.draw_bodypose(canvas, body)
        
    plt.imshow(canvas[:, :, [2, 1, 0]])
    plt.show()



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class ImageBatch(GraphNode):
    image1: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to batch.')
    image2: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to batch.')
    @classmethod
    def get_node_type(cls): return "comfy.image.ImageBatch"



class ImageCompositeMasked(GraphNode):
    destination: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The destination image.')
    source: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The source image.')
    x: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The x position.')
    y: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The y position.')
    resize_source: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Whether to resize the source.')
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=None, description='The mask to use.')
    @classmethod
    def get_node_type(cls): return "comfy.image.ImageCompositeMasked"



class ImageInvert(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to invert.')
    @classmethod
    def get_node_type(cls): return "comfy.image.ImageInvert"



class ImagePadForOutpaint(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to pad.')
    left: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The padding size on the left side.')
    top: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The padding size on the top side.')
    right: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The padding size on the right side.')
    bottom: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The padding size on the bottom side.')
    feathering: int | GraphNode | tuple[GraphNode, str] = Field(default=40, description='The feathering value for softening the edges of the padding.')
    @classmethod
    def get_node_type(cls): return "comfy.image.ImagePadForOutpaint"



class LoadImage(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='The image to load.')
    upload: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='unused')
    @classmethod
    def get_node_type(cls): return "comfy.image.LoadImage"


from nodetool.nodes.comfy.image import ColorChannel

class LoadImageMask(GraphNode):
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to load.')
    channel: ColorChannel | GraphNode | tuple[GraphNode, str] = Field(default=ColorChannel('alpha'), description='The color channel to use.')
    @classmethod
    def get_node_type(cls): return "comfy.image.LoadImageMask"



class PreviewImage(GraphNode):
    images: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to save.')
    filename_prefix: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The prefix for the filename where the image will be saved.')
    @classmethod
    def get_node_type(cls): return "comfy.image.PreviewImage"



class SaveImage(GraphNode):
    images: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to save.')
    filename_prefix: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The prefix for the filename where the image will be saved.')
    @classmethod
    def get_node_type(cls): return "comfy.image.SaveImage"





import warnings
import cv2
import numpy as np
from PIL import Image
from controlnet_aux.util import HWC3, resize_image_with_pad, common_input_validate, HWC3

#Not to be confused with "scribble" from HED. That is "fake scribble" which is more accurate and less picky than this.
class ScribbleDetector:
    def __call__(self, input_image=None, detect_resolution=512, output_type=None, upscale_method="INTER_AREA", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        detected_map = np.zeros_like(input_image, dtype=np.uint8)
        detected_map[np.min(input_image, axis=2) < 127] = 255
        detected_map = 255 - detected_map

        detected_map = remove_pad(detected_map)
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map

class ScribbleXDog_Detector:
    def __call__(self, input_image=None, detect_resolution=512, thr_a=32, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        g1 = cv2.GaussianBlur(input_image.astype(np.float32), (0, 0), 0.5)
        g2 = cv2.GaussianBlur(input_image.astype(np.float32), (0, 0), 5.0)
        dog = (255 - np.min(g2 - g1, axis=2)).clip(0, 255).astype(np.uint8)
        result = np.zeros_like(input_image, dtype=np.uint8)
        result[2 * (255 - dog) > thr_a] = 255
        #result = 255 - result

        detected_map = HWC3(remove_pad(result))
        
        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)
            
        return detected_map


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class CLIPVisionLoader(GraphNode):
    clip_name: CLIPVisionFile | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVisionFile(type='comfy.clip_vision_file', name=''), description='The name of the CLIP vision model to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.CLIPVisionLoader"



class CheckpointLoader(GraphNode):
    ckpt_name: CheckpointFile | GraphNode | tuple[GraphNode, str] = Field(default=CheckpointFile(type='comfy.checkpoint_file', name=''), description='The checkpoint to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.CheckpointLoader"



class CheckpointLoaderSimple(GraphNode):
    ckpt_name: CheckpointFile | GraphNode | tuple[GraphNode, str] = Field(default=CheckpointFile(type='comfy.checkpoint_file', name=''), description='The checkpoint to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.CheckpointLoaderSimple"



class ControlNetLoader(GraphNode):
    control_net_name: ControlNetFile | GraphNode | tuple[GraphNode, str] = Field(default=ControlNetFile(type='comfy.control_net_file', name=''), description='The filename of the control net to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.ControlNetLoader"



class GLIGENLoader(GraphNode):
    gligen_name: GLIGENFile | GraphNode | tuple[GraphNode, str] = Field(default=GLIGENFile(type='comfy.gligen_file', name=''), description='The GLIGEN checkpoint to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.GLIGENLoader"



class UpscaleModelLoader(GraphNode):
    model_name: UpscaleModelFile | GraphNode | tuple[GraphNode, str] = Field(default=UpscaleModelFile(type='comfy.upscale_model_file', name=''), description='The filename of the upscale model to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.UpscaleModelLoader"



class unCLIPCheckpointLoader(GraphNode):
    ckpt_name: unCLIPFile | GraphNode | tuple[GraphNode, str] = Field(default=unCLIPFile(type='comfy.unclip_file', name=''), description='The checkpoint to load.')
    @classmethod
    def get_node_type(cls): return "comfy.loaders.unCLIPCheckpointLoader"





# ------------------------------------------------------------------------------
# Reference: https://github.com/facebookresearch/Mask2Former/blob/main/mask2former/maskformer_model.py
# Modified by Jitesh Jain (https://github.com/praeclarumjj3)
# ------------------------------------------------------------------------------

from typing import Tuple

import torch
from torch import nn
from torch.nn import functional as F

from custom_detectron2.config import configurable
from custom_detectron2.data import MetadataCatalog
from custom_detectron2.modeling import META_ARCH_REGISTRY, build_backbone, build_sem_seg_head
from custom_detectron2.modeling.backbone import Backbone
from custom_detectron2.modeling.postprocessing import sem_seg_postprocess
from custom_detectron2.structures import Boxes, ImageList, Instances, BitMasks
from custom_detectron2.utils.memory import retry_if_cuda_oom

from .modeling.matcher import HungarianMatcher
from einops import rearrange
from .modeling.transformer_decoder.text_transformer import TextTransformer
from .modeling.transformer_decoder.oneformer_transformer_decoder import MLP
from custom_oneformer.data.tokenizer import SimpleTokenizer, Tokenize

@META_ARCH_REGISTRY.register()
class OneFormer(nn.Module):
    """
    Main class for mask classification semantic segmentation architectures.
    """

    @configurable
    def __init__(
        self,
        *,
        backbone: Backbone,
        sem_seg_head: nn.Module,
        task_mlp: nn.Module,
        text_encoder: nn.Module,
        text_projector: nn.Module,
        prompt_ctx: nn.Embedding,
        num_queries: int,
        object_mask_threshold: float,
        overlap_threshold: float,
        metadata,
        size_divisibility: int,
        sem_seg_postprocess_before_inference: bool,
        pixel_mean: Tuple[float],
        pixel_std: Tuple[float],
        # inference
        semantic_on: bool,
        panoptic_on: bool,
        instance_on: bool,
        detection_on: bool,
        test_topk_per_image: int,
        task_seq_len: int,
        max_seq_len: int,
        is_demo: bool,
    ):
        """
        Args:
            backbone: a backbone module, must follow detectron2's backbone interface
            sem_seg_head: a module that predicts semantic segmentation from backbone features
            criterion: a module that defines the loss
            num_queries: int, number of queries
            object_mask_threshold: float, threshold to filter query based on classification score
                for panoptic segmentation inference
            overlap_threshold: overlap threshold used in general inference for panoptic segmentation
            metadata: dataset meta, get `thing` and `stuff` category names for panoptic
                segmentation inference
            size_divisibility: Some backbones require the input height and width to be divisible by a
                specific integer. We can use this to override such requirement.
            sem_seg_postprocess_before_inference: whether to resize the prediction back
                to original input size before semantic segmentation inference or after.
                For high-resolution dataset like Mapillary, resizing predictions before
                inference will cause OOM error.
            pixel_mean, pixel_std: list or tuple with #channels element, representing
                the per-channel mean and std to be used to normalize the input image
            semantic_on: bool, whether to output semantic segmentation prediction
            instance_on: bool, whether to output instance segmentation prediction
            panoptic_on: bool, whether to output panoptic segmentation prediction
            test_topk_per_image: int, instance segmentation parameter, keep topk instances per image
        """
        super().__init__()
        self.backbone = backbone
        self.sem_seg_head = sem_seg_head
        self.task_mlp = task_mlp
        self.text_encoder = text_encoder
        self.text_projector = text_projector
        self.prompt_ctx = prompt_ctx
        self.num_queries = num_queries
        self.overlap_threshold = overlap_threshold
        self.object_mask_threshold = object_mask_threshold
        self.metadata = metadata
        if size_divisibility < 0:
            # use backbone size_divisibility if not set
            size_divisibility = self.backbone.size_divisibility
        self.size_divisibility = size_divisibility
        self.sem_seg_postprocess_before_inference = sem_seg_postprocess_before_inference
        self.register_buffer("pixel_mean", torch.Tensor(pixel_mean).view(-1, 1, 1), False)
        self.register_buffer("pixel_std", torch.Tensor(pixel_std).view(-1, 1, 1), False)

        # additional args
        self.semantic_on = semantic_on
        self.instance_on = instance_on
        self.panoptic_on = panoptic_on
        self.detection_on = detection_on
        self.test_topk_per_image = test_topk_per_image

        self.text_tokenizer = Tokenize(SimpleTokenizer(), max_seq_len=max_seq_len)
        self.task_tokenizer = Tokenize(SimpleTokenizer(), max_seq_len=task_seq_len)
        self.is_demo = is_demo

        self.thing_indices = [k for k in self.metadata.thing_dataset_id_to_contiguous_id.keys()]

        if not self.semantic_on:
            assert self.sem_seg_postprocess_before_inference

    @classmethod
    def from_config(cls, cfg):
        backbone = build_backbone(cfg)
        sem_seg_head = build_sem_seg_head(cfg, backbone.output_shape())

        if cfg.MODEL.IS_TRAIN:
            text_encoder = TextTransformer(context_length=cfg.MODEL.TEXT_ENCODER.CONTEXT_LENGTH,
                                    width=cfg.MODEL.TEXT_ENCODER.WIDTH,
                                    layers=cfg.MODEL.TEXT_ENCODER.NUM_LAYERS,
                                    vocab_size=cfg.MODEL.TEXT_ENCODER.VOCAB_SIZE)
            text_projector = MLP(text_encoder.width, cfg.MODEL.ONE_FORMER.HIDDEN_DIM, 
                                cfg.MODEL.ONE_FORMER.HIDDEN_DIM, cfg.MODEL.TEXT_ENCODER.PROJ_NUM_LAYERS)
            if cfg.MODEL.TEXT_ENCODER.N_CTX > 0:
                prompt_ctx = nn.Embedding(cfg.MODEL.TEXT_ENCODER.N_CTX, cfg.MODEL.TEXT_ENCODER.WIDTH)
            else:
                prompt_ctx = None
        else:
            text_encoder = None
            text_projector = None
            prompt_ctx = None

        task_mlp = MLP(cfg.INPUT.TASK_SEQ_LEN, cfg.MODEL.ONE_FORMER.HIDDEN_DIM,
                        cfg.MODEL.ONE_FORMER.HIDDEN_DIM, 2)

        # Loss parameters:
        deep_supervision = cfg.MODEL.ONE_FORMER.DEEP_SUPERVISION
        no_object_weight = cfg.MODEL.ONE_FORMER.NO_OBJECT_WEIGHT

        # loss weights
        class_weight = cfg.MODEL.ONE_FORMER.CLASS_WEIGHT
        dice_weight = cfg.MODEL.ONE_FORMER.DICE_WEIGHT
        mask_weight = cfg.MODEL.ONE_FORMER.MASK_WEIGHT
        contrastive_weight = cfg.MODEL.ONE_FORMER.CONTRASTIVE_WEIGHT
        
        # building criterion
        matcher = HungarianMatcher(
            cost_class=class_weight,
            cost_mask=mask_weight,
            cost_dice=dice_weight,
            num_points=cfg.MODEL.ONE_FORMER.TRAIN_NUM_POINTS,
        )

        weight_dict = {"loss_ce": class_weight, "loss_mask": mask_weight, 
                        "loss_dice": dice_weight, "loss_contrastive": contrastive_weight}

        
        if deep_supervision:
            dec_layers = cfg.MODEL.ONE_FORMER.DEC_LAYERS
            aux_weight_dict = {}
            for i in range(dec_layers - 1):
                aux_weight_dict.update({k + f"_{i}": v for k, v in weight_dict.items()})
            weight_dict.update(aux_weight_dict)

        losses = ["labels", "masks", "contrastive"]

        return {
            "backbone": backbone,
            "sem_seg_head": sem_seg_head,
            "task_mlp": task_mlp,
            "prompt_ctx": prompt_ctx,
            "text_encoder": text_encoder,
            "text_projector": text_projector,
            "num_queries": cfg.MODEL.ONE_FORMER.NUM_OBJECT_QUERIES,
            "object_mask_threshold": cfg.MODEL.TEST.OBJECT_MASK_THRESHOLD,
            "overlap_threshold": cfg.MODEL.TEST.OVERLAP_THRESHOLD,
            "metadata": MetadataCatalog.get(cfg.DATASETS.TRAIN[0]),
            "size_divisibility": cfg.MODEL.ONE_FORMER.SIZE_DIVISIBILITY,
            "sem_seg_postprocess_before_inference": (
                cfg.MODEL.TEST.SEM_SEG_POSTPROCESSING_BEFORE_INFERENCE
                or cfg.MODEL.TEST.PANOPTIC_ON
                or cfg.MODEL.TEST.INSTANCE_ON
            ),
            "pixel_mean": cfg.MODEL.PIXEL_MEAN,
            "pixel_std": cfg.MODEL.PIXEL_STD,
            # inference
            "semantic_on": cfg.MODEL.TEST.SEMANTIC_ON,
            "instance_on": cfg.MODEL.TEST.INSTANCE_ON,
            "panoptic_on": cfg.MODEL.TEST.PANOPTIC_ON,
            "detection_on": cfg.MODEL.TEST.DETECTION_ON,
            "test_topk_per_image": cfg.TEST.DETECTIONS_PER_IMAGE,
            "task_seq_len": cfg.INPUT.TASK_SEQ_LEN,
            "max_seq_len": cfg.INPUT.MAX_SEQ_LEN,
            "is_demo": cfg.MODEL.IS_DEMO,
        }

    @property
    def device(self):
        return self.pixel_mean.device

    def encode_text(self, text):
        assert text.ndim in [2, 3], text.ndim
        b = text.shape[0]
        squeeze_dim = False
        num_text = 1
        if text.ndim == 3:
            num_text = text.shape[1]
            text = rearrange(text, 'b n l -> (b n) l', n=num_text)
            squeeze_dim = True

        # [B, C]
        x = self.text_encoder(text)

        text_x = self.text_projector(x)

        if squeeze_dim:
            text_x = rearrange(text_x, '(b n) c -> b n c', n=num_text)
            if self.prompt_ctx is not None:
                text_ctx = self.prompt_ctx.weight.unsqueeze(0).repeat(text_x.shape[0], 1, 1)
                text_x = torch.cat([text_x, text_ctx], dim=1)
        
        return {"texts": text_x}
    
    def forward(self, batched_inputs):
        """
        Args:
            batched_inputs: a list, batched outputs of :class:`DatasetMapper`.
                Each item in the list contains the inputs for one image.
                For now, each item in the list is a dict that contains:
                   * "image": Tensor, image in (C, H, W) format.
                   * "instances": per-region ground truth
                   * Other information that's included in the original dicts, such as:
                     "height", "width" (int): the output resolution of the model (may be different
                     from input resolution), used in inference.
        Returns:
            list[dict]:
                each dict has the results for one image. The dict contains the following keys:
                * "sem_seg":
                    A Tensor that represents the
                    per-pixel segmentation prediced by the head.
                    The prediction has shape KxHxW that represents the logits of
                    each class for each pixel.
                * "panoptic_seg":
                    A tuple that represent panoptic output
                    panoptic_seg (Tensor): of shape (height, width) where the values are ids for each segment.
                    segments_info (list[dict]): Describe each segment in `panoptic_seg`.
                        Each dict contains keys "id", "category_id", "isthing".
        """
        images = [x["image"].to(self.device) for x in batched_inputs]
        images = [(x - self.pixel_mean) / self.pixel_std for x in images]
        images = ImageList.from_tensors(images, self.size_divisibility)

        tasks = torch.cat([self.task_tokenizer(x["task"]).to(self.device).unsqueeze(0) for x in batched_inputs], dim=0)
        tasks = self.task_mlp(tasks.float())

        features = self.backbone(images.tensor)
        outputs = self.sem_seg_head(features, tasks)

        if self.training:
            texts = torch.cat([self.text_tokenizer(x["text"]).to(self.device).unsqueeze(0) for x in batched_inputs], dim=0)
            texts_x = self.encode_text(texts)

            outputs = {**outputs, **texts_x}

            # mask classification target
            if "instances" in batched_inputs[0]:
                gt_instances = [x["instances"].to(self.device) for x in batched_inputs]
                targets = self.prepare_targets(gt_instances, images)
            else:
                targets = None

            # bipartite matching-based loss
            losses = self.criterion(outputs, targets)

            for k in list(losses.keys()):
                if k in self.criterion.weight_dict:
                    losses[k] *= self.criterion.weight_dict[k]
                else:
                    # remove this loss if not specified in `weight_dict`
                    losses.pop(k)
            return losses
        else:
            mask_cls_results = outputs["pred_logits"]
            mask_pred_results = outputs["pred_masks"]
            # upsample masks
            mask_pred_results = F.interpolate(
                mask_pred_results,
                size=(images.tensor.shape[-2], images.tensor.shape[-1]),
                mode="bilinear",
                align_corners=False,
            )

            del outputs

            processed_results = []
            for i, data in enumerate(zip(
                mask_cls_results, mask_pred_results, batched_inputs, images.image_sizes
            )):
                mask_cls_result, mask_pred_result, input_per_image, image_size = data
                height = input_per_image.get("height", image_size[0])
                width = input_per_image.get("width", image_size[1])
                processed_results.append({})

                if self.sem_seg_postprocess_before_inference:
                    mask_pred_result = retry_if_cuda_oom(sem_seg_postprocess)(
                        mask_pred_result, image_size, height, width
                    )
                    mask_cls_result = mask_cls_result.to(mask_pred_result)

                # semantic segmentation inference
                if self.semantic_on:
                    r = retry_if_cuda_oom(self.semantic_inference)(mask_cls_result, mask_pred_result)
                    if not self.sem_seg_postprocess_before_inference:
                        r = retry_if_cuda_oom(sem_seg_postprocess)(r, image_size, height, width)
                    processed_results[-1]["sem_seg"] = r

                # panoptic segmentation inference
                if self.panoptic_on:
                    panoptic_r = retry_if_cuda_oom(self.panoptic_inference)(mask_cls_result, mask_pred_result)
                    processed_results[-1]["panoptic_seg"] = panoptic_r
                
                # instance segmentation inference
                if self.instance_on:
                    instance_r = retry_if_cuda_oom(self.instance_inference)(mask_cls_result, mask_pred_result)
                    processed_results[-1]["instances"] = instance_r

                if self.detection_on:
                    bbox_r = retry_if_cuda_oom(self.instance_inference)(mask_cls_result, mask_pred_result)
                    processed_results[-1]["box_instances"] = bbox_r

            return processed_results

    def prepare_targets(self, targets, images):
        h_pad, w_pad = images.tensor.shape[-2:]
        new_targets = []
        for targets_per_image in targets:
            # pad gt
            gt_masks = targets_per_image.gt_masks
            padded_masks = torch.zeros((gt_masks.shape[0], h_pad, w_pad), dtype=gt_masks.dtype, device=gt_masks.device)
            padded_masks[:, : gt_masks.shape[1], : gt_masks.shape[2]] = gt_masks
            new_targets.append(
                {
                    "labels": targets_per_image.gt_classes,
                    "masks": padded_masks,
                }
            )
        return new_targets

    def semantic_inference(self, mask_cls, mask_pred):
        mask_cls = F.softmax(mask_cls, dim=-1)[..., :-1]
        mask_pred = mask_pred.sigmoid()
        semseg = torch.einsum("qc,qhw->chw", mask_cls, mask_pred)
        return semseg

    def panoptic_inference(self, mask_cls, mask_pred):
        scores, labels = F.softmax(mask_cls, dim=-1).max(-1)
        mask_pred = mask_pred.sigmoid()

        keep = labels.ne(self.sem_seg_head.num_classes) & (scores > self.object_mask_threshold)
        cur_scores = scores[keep]
        cur_classes = labels[keep]
        cur_masks = mask_pred[keep]
        cur_mask_cls = mask_cls[keep]
        cur_mask_cls = cur_mask_cls[:, :-1]

        cur_prob_masks = cur_scores.view(-1, 1, 1) * cur_masks

        h, w = cur_masks.shape[-2:]
        panoptic_seg = torch.zeros((h, w), dtype=torch.int32, device=cur_masks.device)
        segments_info = []

        current_segment_id = 0

        if cur_masks.shape[0] == 0:
            # We didn't detect any mask :(
            return panoptic_seg, segments_info
        else:
            # take argmax
            cur_mask_ids = cur_prob_masks.argmax(0)
            stuff_memory_list = {}
            for k in range(cur_classes.shape[0]):
                pred_class = cur_classes[k].item()
                isthing = pred_class in self.metadata.thing_dataset_id_to_contiguous_id.values()
                mask_area = (cur_mask_ids == k).sum().item()
                original_area = (cur_masks[k] >= 0.5).sum().item()
                mask = (cur_mask_ids == k) & (cur_masks[k] >= 0.5)

                if mask_area > 0 and original_area > 0 and mask.sum().item() > 0:
                    if mask_area / original_area < self.overlap_threshold:
                        continue

                    # merge stuff regions
                    if not isthing:
                        if int(pred_class) in stuff_memory_list.keys():
                            panoptic_seg[mask] = stuff_memory_list[int(pred_class)]
                            continue
                        else:
                            stuff_memory_list[int(pred_class)] = current_segment_id + 1

                    current_segment_id += 1
                    panoptic_seg[mask] = current_segment_id

                    segments_info.append(
                        {
                            "id": current_segment_id,
                            "isthing": bool(isthing),
                            "category_id": int(pred_class),
                        }
                    )

            return panoptic_seg, segments_info

    def instance_inference(self, mask_cls, mask_pred):
        # mask_pred is already processed to have the same shape as original input
        image_size = mask_pred.shape[-2:]

        # [Q, K]
        scores = F.softmax(mask_cls, dim=-1)[:, :-1]
        labels = torch.arange(self.sem_seg_head.num_classes, device=self.device).unsqueeze(0).repeat(self.num_queries, 1).flatten(0, 1)
        
        # scores_per_image, topk_indices = scores.flatten(0, 1).topk(self.num_queries, sorted=False)
        scores_per_image, topk_indices = scores.flatten(0, 1).topk(self.test_topk_per_image, sorted=False)
        labels_per_image = labels[topk_indices]

        topk_indices = topk_indices // self.sem_seg_head.num_classes
        # mask_pred = mask_pred.unsqueeze(1).repeat(1, self.sem_seg_head.num_classes, 1).flatten(0, 1)
        mask_pred = mask_pred[topk_indices]

        # Only consider scores with confidence over [self.object_mask_threshold] for demo
        if self.is_demo:
            keep = scores_per_image > self.object_mask_threshold
            scores_per_image = scores_per_image[keep]
            labels_per_image = labels_per_image[keep]
            mask_pred = mask_pred[keep]

        # if this is panoptic segmentation, we only keep the "thing" classes
        if self.panoptic_on:
            keep = torch.zeros_like(scores_per_image).bool()
            for i, lab in enumerate(labels_per_image):
                keep[i] = lab in self.metadata.thing_dataset_id_to_contiguous_id.values()

            scores_per_image = scores_per_image[keep]
            labels_per_image = labels_per_image[keep]
            mask_pred = mask_pred[keep]
        
        if 'ade20k' in self.metadata.name:
            for i in range(labels_per_image.shape[0]):
                labels_per_image[i] = self.thing_indices.index(labels_per_image[i].item())

        result = Instances(image_size)
        # mask (before sigmoid)
        result.pred_masks = (mask_pred > 0).float()
        if self.detection_on:
            # Uncomment the following to get boxes from masks (this is slow)
            result.pred_boxes = BitMasks(mask_pred > 0).get_bounding_boxes()
        else:
            result.pred_boxes = Boxes(torch.zeros(mask_pred.size(0), 4))

        # calculate average mask prob
        mask_scores_per_image = (mask_pred.sigmoid().flatten(1) * result.pred_masks.flatten(1)).sum(1) / (result.pred_masks.flatten(1).sum(1) + 1e-6)
        result.scores = scores_per_image * mask_scores_per_image
        result.pred_classes = labels_per_image
        return result


from enum import EnumMeta
import inspect
from typing import Any, Union, get_args, get_origin
from types import UnionType
from collections.abc import Generator, AsyncGenerator


def is_generator_type(t):
    """
    Check if a type is a generator.

    Args:
        t: The type to check.

    Returns:
        True if the type is a generator, False otherwise.
    """
    return get_origin(t) is Generator


def is_async_generator_type(t):
    """
    Check if a type is an async generator.

    Args:
        t: The type to check.

    Returns:
        True if the type is an async generator, False otherwise.
    """
    return get_origin(t) is AsyncGenerator


def is_optional_type(t):
    """
    Check if a type is an optional type.

    Args:
        t: The type to check.

    Returns:
        True if the type is an optional type, False otherwise.
    """
    return (
        get_origin(t) is Union and len(get_args(t)) == 2 and type(None) in get_args(t)
    )


def is_enum_type(t):
    """
    Check if a type is an enum.

    Args:
        t: The type to check.

    Returns:
        True if the type is an enum, False otherwise.
    """
    return isinstance(t, EnumMeta)


def is_union_type(t):
    """
    Check if a type is a union.

    Args:
        t: The type to check.

    Returns:
        True if the type is a union, False otherwise.
    """
    return get_origin(t) is Union or isinstance(t, UnionType)


def is_list_type(t):
    """
    Check if a type is a list.

    Args:
        t: The type to check.

    Returns:
        True if the type is a list, False otherwise.
    """
    return t is list or get_origin(t) is list


def is_dict_type(t):
    """
    Check if a type is a dictionary.

    Args:
        t: The type to check.

    Returns:
        True if the type is a dictionary, False otherwise.
    """
    return t is dict or get_origin(t) is dict


def is_number_type(t):
    """
    Check if a type is a number.

    Args:
        t: The type to check.

    Returns:
        True if the type is a number, False otherwise.
    """
    from nodetool.metadata.types import Tensor

    return is_union_type(t) and all(
        [arg in [int, float, Tensor] for arg in get_args(t)]
    )


def is_class(obj: Any) -> bool:
    return inspect.isclass(obj)



from enum import Enum
from pydantic import Field
from nodetool.metadata.types import ImageTensor, Mask
from nodetool.common.comfy_node import ComfyNode
from nodetool.nodes.comfy.image import ColorChannel


class ChannelEnum(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    ALPHA = "alpha"


class OperationEnum(str, Enum):
    MULTIPLY = "multiply"
    ADD = "add"
    SUBTRACT = "subtract"
    AND = "and"
    OR = "or"
    XOR = "xor"


class MaskToImage(ComfyNode):
    """
    The Convert Mask to Image node can be used to convert a mask to a grey scale image.
    """

    mask: Mask = Field(default=Mask(), description="The mask to convert.")

    @classmethod
    def get_title(cls):
        return "Convert Mask to Image"

    @classmethod
    def return_type(cls):
        return {"image": ImageTensor}


class ImageToMask(ComfyNode):
    """
    The Convert Image yo Mask node can be used to convert a specific channel of an image into a mask.
    """

    image: ImageTensor = Field(
        default=ImageTensor(), description="The image to extract the mask."
    )
    channel: ChannelEnum = Field(
        default=ChannelEnum.RED, description="The channel to use for the mask."
    )

    @classmethod
    def get_title(cls):
        return "Convert Image to Mask"

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class ImageColorToMask(ComfyNode):
    image: ImageTensor = Field(
        default=ImageTensor(), description="The image to extract the color mask."
    )
    color: int = Field(default=0, description="The color to use for the mask.")

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class SolidMask(ComfyNode):
    """
    The Solid Mask node can be used to create a solid masking containing a single value.
    """

    value: float = Field(default=1.0, description="The value for the solid mask.")
    width: int = Field(default=512, description="Width of the solid mask.")
    height: int = Field(default=512, description="Height of the solid mask.")

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class InvertMask(ComfyNode):
    """
    The Invert Mask node can be used to invert a mask.
    """

    mask: Mask = Field(default=Mask(), description="The mask to invert.")

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class CropMask(ComfyNode):
    """
    The Crop Mask node can be used to crop a mask to a new shape.
    """

    mask: Mask = Field(default=Mask(), description="The mask to crop.")
    x: int = Field(default=0, description="The x position for cropping.")
    y: int = Field(default=0, description="The y position for cropping.")
    width: int = Field(default=512, description="Width of the crop.")
    height: int = Field(default=512, description="Height of the crop.")

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class MaskComposite(ComfyNode):
    """
    The Mask Composite node can be used to paste one mask into another.
    """

    destination: Mask = Field(default=Mask(), description="The destination mask.")
    source: Mask = Field(default=Mask(), description="The source mask.")
    x: int = Field(default=0, description="The x position.")
    y: int = Field(default=0, description="The y position.")
    operation: OperationEnum = Field(
        default=OperationEnum.MULTIPLY, description="The operation to use."
    )

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class FeatherMask(ComfyNode):
    """
    The Feather Mask node can be used to feather a mask.
    """

    mask: Mask = Field(default=Mask(), description="The mask to feather.")
    left: int = Field(default=0, description="Feather amount on the left.")
    top: int = Field(default=0, description="Feather amount on the top.")
    right: int = Field(default=0, description="Feather amount on the right.")
    bottom: int = Field(default=0, description="Feather amount on the bottom.")

    @classmethod
    def return_type(cls):
        return {"mask": Mask}


class GrowMask(ComfyNode):
    mask: Mask = Field(default=Mask(), description="The mask to grow.")
    expand: int = Field(default=0, description="The amount to expand the mask.")
    tapered_corners: bool = Field(
        default=True, description="Whether to taper the corners."
    )

    @classmethod
    def return_type(cls):
        return {"mask": Mask}



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class FlipSigmas(GraphNode):
    sigmas: Sigmas | GraphNode | tuple[GraphNode, str] = Field(default=Sigmas(type='comfy.sigmas'), description='The array of sigmas to flip.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.sigmas.FlipSigmas"



class SplitSigmas(GraphNode):
    sigmas: Sigmas | GraphNode | tuple[GraphNode, str] = Field(default=Sigmas(type='comfy.sigmas'), description='The array of sigmas to split.')
    step: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The specific step at which to split the sigmas array.')
    @classmethod
    def get_node_type(cls): return "comfy.sampling.sigmas.SplitSigmas"





# Create the DynamoDB table
from nodetool.common.environment import Environment

from nodetool.models.asset import Asset
from nodetool.models.job import Job
from nodetool.models.message import Message
from nodetool.models.prediction import Prediction
from nodetool.models.task import Task
from nodetool.models.thread import Thread
from nodetool.models.user import User
from nodetool.models.workflow import Workflow

log = Environment.get_logger()

models = [Asset, Job, Message, Prediction, Task, Thread, User, Workflow]


def create_all_tables():
    for model in models:
        model.create_table()


def drop_all_tables():
    for model in models:
        model.drop_table()



#!/usr/bin/env python

from fastapi import APIRouter
from nodetool.api.types.job import (
    JobUpdate,
)
from nodetool.api.types.prediction import Prediction
from nodetool.metadata.node_metadata import NodeMetadata
from nodetool.metadata.types import (
    AssetRef,
    NodeRef,
    WorkflowRef,
    AudioRef,
    DataframeRef,
    FolderRef,
    ImageRef,
    ModelRef,
    Tensor,
    TextRef,
    VideoRef,
)
from nodetool.workflows.base_node import get_registered_node_classes
from nodetool.workflows.types import (
    WorkflowUpdate,
)

from nodetool.common.environment import Environment
from nodetool.metadata.types import (
    FileRef,
)
from nodetool.workflows.types import NodeProgress, NodeUpdate

log = Environment.get_logger()
router = APIRouter(prefix="/api/nodes", tags=["nodes"])


# This is a dummy type that contains all property types and Websocket types.
UnionType = (
    AssetRef
    | AudioRef
    | DataframeRef
    | FolderRef
    | ImageRef
    | Tensor
    | VideoRef
    | FileRef
    | ModelRef
    | TextRef
    | WorkflowRef
    | NodeRef
    | Prediction
    | JobUpdate
    | NodeUpdate
    | NodeProgress
    | WorkflowUpdate
    | dict
)


@router.get("/dummy")
async def dummy() -> UnionType:
    """
    Returns a dummy node.
    """
    return {"hello": "world"}


@router.get("/metadata")
async def metadata() -> list[NodeMetadata]:
    """
    Returns a list of all node metadata.
    """
    import nodetool.nodes

    return [node_class.metadata() for node_class in get_registered_node_classes()]



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.huggingface.audio.speech_recognition import ModelId

class AutomaticSpeechRecognition(GraphNode):
    model: ModelId | GraphNode | tuple[GraphNode, str] = Field(default=ModelId('openai/whisper-large-v3'), description='The model ID to use for the speech recognition')
    audio: AudioRef | GraphNode | tuple[GraphNode, str] = Field(default=AudioRef(type='audio', uri='', asset_id=None), description='The input audio to transcribe')
    @classmethod
    def get_node_type(cls): return "huggingface.audio.speech_recognition.AutomaticSpeechRecognition"





import warnings

import cv2
import numpy as np
from PIL import Image
import random

from controlnet_aux.util import HWC3, common_input_validate, img2mask, make_noise_disk, resize_image_with_pad


class ContentShuffleDetector:
    def __call__(self, input_image, h=None, w=None, f=None, detect_resolution=512, output_type="pil", upscale_method="INTER_CUBIC", **kwargs):
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        input_image, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)

        H, W, C = input_image.shape
        if h is None:
            h = H
        if w is None:
            w = W
        if f is None:
            f = 256
        x = make_noise_disk(h, w, 1, f) * float(W - 1)
        y = make_noise_disk(h, w, 1, f) * float(H - 1)
        flow = np.concatenate([x, y], axis=2).astype(np.float32)
        detected_map = cv2.remap(input_image, flow, None, cv2.INTER_LINEAR)
        detected_map = remove_pad(detected_map)

        if output_type == "pil":
            detected_map = Image.fromarray(detected_map)

        return detected_map


class ColorShuffleDetector:
    def __call__(self, img):
        H, W, C = img.shape
        F = np.random.randint(64, 384)
        A = make_noise_disk(H, W, 3, F)
        B = make_noise_disk(H, W, 3, F)
        C = (A + B) / 2.0
        A = (C + (A - C) * 3.0).clip(0, 1)
        B = (C + (B - C) * 3.0).clip(0, 1)
        L = img.astype(np.float32) / 255.0
        Y = A * L + B * (1 - L)
        Y -= np.min(Y, axis=(0, 1), keepdims=True)
        Y /= np.maximum(np.max(Y, axis=(0, 1), keepdims=True), 1e-5)
        Y *= 255.0
        return Y.clip(0, 255).astype(np.uint8)


class GrayDetector:
    def __call__(self, img):
        eps = 1e-5
        X = img.astype(np.float32)
        r, g, b = X[:, :, 0], X[:, :, 1], X[:, :, 2]
        kr, kg, kb = [random.random() + eps for _ in range(3)]
        ks = kr + kg + kb
        kr /= ks
        kg /= ks
        kb /= ks
        Y = r * kr + g * kg + b * kb
        Y = np.stack([Y] * 3, axis=2)
        return Y.clip(0, 255).astype(np.uint8)


class DownSampleDetector:
    def __call__(self, img, level=3, k=16.0):
        h = img.astype(np.float32)
        for _ in range(level):
            h += np.random.normal(loc=0.0, scale=k, size=h.shape)
            h = cv2.pyrDown(h)
        for _ in range(level):
            h = cv2.pyrUp(h)
            h += np.random.normal(loc=0.0, scale=k, size=h.shape)
        return h.clip(0, 255).astype(np.uint8)


class Image2MaskShuffleDetector:
    def __init__(self, resolution=(640, 512)):
        self.H, self.W = resolution

    def __call__(self, img):
        m = img2mask(img, self.H, self.W)
        m *= 255.0
        return m.clip(0, 255).astype(np.uint8)



#!/usr/bin/env python

from enum import Enum
import re

from fastapi import APIRouter
from pydantic import BaseModel, validator
from authlib.integrations.requests_client import OAuth2Session

from nodetool.common.environment import Environment
from nodetool.models.user import User

log = Environment.get_logger()
router = APIRouter(prefix="/api/auth", tags=["auth"])


class TokenRequest(BaseModel):
    token: str


class TokenResponse(BaseModel):
    valid: bool


class OAuthLoginResponse(BaseModel):
    url: str
    state: str


class OAuthProvider(str, Enum):
    google = "google"
    facebook = "facebook"


class OAuthLoginRequest(BaseModel):
    redirect_uri: str
    provider: OAuthProvider


class OAuthAuthorizeRequest(BaseModel):
    provider: OAuthProvider
    state: str
    authorization_response: str
    redirect_uri: str


OAUTH_AUTHORIZATION_ENDPOINT = {
    OAuthProvider.google: "https://accounts.google.com/o/oauth2/v2/auth",
}

OAUTH_TOKEN_ENDPOINT = {
    OAuthProvider.google: "https://oauth2.googleapis.com/token",
}


def get_user_email(client: OAuth2Session, provider: OAuthProvider) -> str:
    if provider == OAuthProvider.google:
        return client.get("https://www.googleapis.com/oauth2/v1/userinfo").json()[
            "email"
        ]
    else:
        raise ValueError(f"Unknown provider: {provider}")


@router.post("/oauth/login")
async def oauth_login(login: OAuthLoginRequest) -> OAuthLoginResponse:
    authorization_endpoint = OAUTH_AUTHORIZATION_ENDPOINT[login.provider]
    client = Environment.get_google_oauth2_session()
    uri, state = client.create_authorization_url(
        authorization_endpoint, redirect_uri=login.redirect_uri
    )
    return OAuthLoginResponse(url=uri, state=state)


@router.post("/oauth/callback")
async def oauth_callback(auth: OAuthAuthorizeRequest) -> User:
    token_endpoint = OAUTH_TOKEN_ENDPOINT[auth.provider]
    client = Environment.get_google_oauth2_session()
    res = client.fetch_token(
        token_endpoint,
        state=auth.state,
        authorization_response=auth.authorization_response,
        redirect_uri=auth.redirect_uri,
        grant_type="authorization_code",
    )

    email = get_user_email(client, auth.provider)

    # email is trusted at this point
    user, created = User.create(email, verified=True)
    if created:
        log.info(f"Created user: {user.email}")
    else:
        log.info(f"User already exists: {user.email}")

    return User(email=user.email, id=user.id, auth_token=user.auth_token)


class AuthRequest(BaseModel):
    token: str


@router.post("/verify")
async def verify(auth: AuthRequest) -> TokenResponse:
    user = User.find_by_auth_token(auth.token)
    if user:
        return TokenResponse(valid=True)
    return TokenResponse(valid=False)



import os

import cv2
import numpy as np
import torch
from einops import rearrange
from PIL import Image

from controlnet_aux.util import HWC3, common_input_validate, resize_image_with_pad, custom_hf_download, HF_MODEL_NAME
from .api import MiDaSInference


class MidasDetector:
    def __init__(self, model):
        self.model = model
        
    @classmethod
    def from_pretrained(cls, pretrained_model_or_path=HF_MODEL_NAME, model_type="dpt_hybrid", filename="dpt_hybrid-midas-501f0c75.pt"):
        subfolder = "annotator/ckpts" if pretrained_model_or_path == "lllyasviel/ControlNet" else ''
        model_path = custom_hf_download(pretrained_model_or_path, filename, subfolder=subfolder)
        model = MiDaSInference(model_type=model_type, model_path=model_path)
        return cls(model)
        

    def to(self, device):
        self.model.to(device)
        return self
    
    def __call__(self, input_image, a=np.pi * 2.0, bg_th=0.1, depth_and_normal=False, detect_resolution=512, output_type=None, upscale_method="INTER_CUBIC", **kwargs):
        device = next(iter(self.model.parameters())).device
        input_image, output_type = common_input_validate(input_image, output_type, **kwargs)
        detected_map, remove_pad = resize_image_with_pad(input_image, detect_resolution, upscale_method)
        image_depth = detected_map
        with torch.no_grad():
            image_depth = torch.from_numpy(image_depth).float()
            image_depth = image_depth.to(device)
            image_depth = image_depth / 127.5 - 1.0
            image_depth = rearrange(image_depth, 'h w c -> 1 c h w')
            depth = self.model(image_depth)[0]

            depth_pt = depth.clone()
            depth_pt -= torch.min(depth_pt)
            depth_pt /= torch.max(depth_pt)
            depth_pt = depth_pt.cpu().numpy()
            depth_image = (depth_pt * 255.0).clip(0, 255).astype(np.uint8)

            if depth_and_normal:
                depth_np = depth.cpu().numpy()
                x = cv2.Sobel(depth_np, cv2.CV_32F, 1, 0, ksize=3)
                y = cv2.Sobel(depth_np, cv2.CV_32F, 0, 1, ksize=3)
                z = np.ones_like(x) * a
                x[depth_pt < bg_th] = 0
                y[depth_pt < bg_th] = 0
                normal = np.stack([x, y, z], axis=2)
                normal /= np.sum(normal ** 2.0, axis=2, keepdims=True) ** 0.5
                normal_image = (normal * 127.5 + 127.5).clip(0, 255).astype(np.uint8)[:, :, ::-1]
        
        depth_image = HWC3(depth_image)
        if depth_and_normal:
            normal_image = HWC3(normal_image)


        depth_image = remove_pad(depth_image)
        if depth_and_normal:
            normal_image = remove_pad(normal_image)
        
        if output_type == "pil":
            depth_image = Image.fromarray(depth_image)
            if depth_and_normal:
                normal_image = Image.fromarray(normal_image)
        
        if depth_and_normal:
            return depth_image, normal_image
        else:
            return depth_image
